#ifndef _LRU_REPLACEMENT_H
#define _LRU_REPLACEMENT_H

#include <DoublyList.h>

	//Forward References
#include <BrushVisitor.h>
#include <ServedItemVisitor.h>
#include <ReplacementPolicy.h>
#include <Cache.h>
#include <CacheEntry.h>
#include <log/log.h>

namespace cache
{
	class IIndex;

	class ReplacementException
	{
	public:
		ReplacementException(){}
	};


	class LRUReplacementValue
	{
	public:
		LRUReplacementValue(){}

		LRUReplacementValue(Cache* cache); 

		LRUReplacementValue(LRUReplacementValue &in)
		{
			this->isFree = in.isFree;
			this->isServed = in.isServed;
			this->nodePtr = in.nodePtr;		
		}
		//Stores if the replacement Value is free
		bool isFree;
		//Stores if the replacement value is available
		bool isServed;
		//Stores a pointer inside the ordered list
		DoublyList<CacheEntry<LRUReplacementValue>*>::Node *nodePtr; 
		//Stores the reference to the messge group
		debug::log::CMessageGroup replacementLogger;
	};


	class LRUReplacement : public ReplacementPolicy

	{
	public:		
		LRUReplacement(Cache *cache,
			CacheEntry<LRUReplacementValue>* buffer,
			unsigned long sizeOfBuffer,
			IIndex *index);		

		~LRUReplacement();   

		ICacheEntry* getVictimItem();

		void newData(ICacheEntry &cacheEntry);

		void servedItem(ICacheEntry& cacheEntry);

		void hitCacheEntry(ICacheEntry& cacheEntry);

		void ifRemainder(bool flag);

		void brushModified(Brush& brush);

		void checkForError();

		ICacheEntry* peek();

		void clearCache();

		void removeEntries(std::list<ICacheEntry*>& listCacheEntry);


	private:
		//Store the Lru list
		DoublyList<CacheEntry<LRUReplacementValue>*>* lruList;
		//Stores the free list
		DoublyList<CacheEntry<LRUReplacementValue>*>* freeList;
		//Stores the not available list
		DoublyList<CacheEntry<LRUReplacementValue>*>* notAvailableList;
		//This ensures that when we refer to lBuffer[i] the increment 
		//to the pointer is defined by the sub type and not the super type.
		//That should be the case.
		CacheEntry<LRUReplacementValue>* lBuffer;	
		//Stores the size of the buffer	
		unsigned long sizeOfBuffer;	
		//Store the reference to the cache
		Cache* cache;
	};

}

#endif //_LRU_REPLACEMENT