#ifndef _RemainderRegion_H
#define _RemainderRegion_H
#include<string>
#include <sstream>

namespace cache
{

	class RemainderRegion
	{
	public:	
		friend class QueryFactory;
	protected:
		RemainderRegion(double e1,double e2,double r)
		{
			this->e1 = e1;
			this->e2 = e2;
			this->r = r;
		}

	public:

		virtual ~RemainderRegion()
		{}

		double getLeftExtent()
		{
			return(this->e1);
		}
		double getRightExtent()
		{
			return(this->e2);
		}

		double getLevelOfDetail()
		{
			return(this->r);
		}
		virtual std::string getPredicate()=0;

	protected:
		double e1;
		double e2;
		double r;
	};

	class AnyRemainderRegion : public RemainderRegion
	{
	public: 
		AnyRemainderRegion(double e1,double e2,double r):RemainderRegion(e1,e2,r)
		{}

		std::string getPredicate()
		{		
			std::ostringstream ostring;
			ostring<<" X <= "<<e2<<" and Y >= "<<e1<<" ";
			return(ostring.str());
		}
		virtual ~AnyRemainderRegion()
		{}
	};

	class LeftAnyRemainderRegion : public AnyRemainderRegion
	{
	public :
		LeftAnyRemainderRegion(double e1,double e2,double r):
		   AnyRemainderRegion(e1,e2,r)
		   {}

		   std::string getPredicate()
		   {
			   std::ostringstream ostring;
			   ostring<<" X < "<<e2<<" and Y >= "<<e1<<" ";				
			   return(ostring.str());
		   }
		   virtual ~LeftAnyRemainderRegion(){}

	};

	class RightAnyRemainderRegion : public AnyRemainderRegion
	{
	public :
		RightAnyRemainderRegion(double e1,double e2,double r):
		   AnyRemainderRegion(e1,e2,r)
		   {}

		   std::string getPredicate()
		   {
			   std::ostringstream ostring;
			   ostring<<" X <= "<<e2<<" and Y > "<<e1<<" ";								
			   return(ostring.str());
		   }

		   virtual ~RightAnyRemainderRegion(){}

	};


	class AllRemainderRegion : public RemainderRegion
	{
	public:
		AllRemainderRegion(double e1,double e2,double r):RemainderRegion(e1,e2,r)
		{	}

		std::string getPredicate()
		{
			std::ostringstream ostring;
			ostring<<"  X >= "<<e1<<" and Y <= "<<e2<<" ";


			return(ostring.str());				
		}

		virtual ~AllRemainderRegion(){}
	};
}
#endif //_RemainderRegion_H