#ifndef _IREPLACEMENT_POLICY_H
#define _IREPLACEMENT_POLICY_H

#include <ServedItemVisitor.h>
#include <BrushVisitor.h>
#include <Cache.h>

namespace cache
{

	//Forward Reference
	class ICacheEntry;


	//Defines an interface to be implemented by each replacement Policy
	class ReplacementPolicy : public BrushVisitor, 
		public ServedItemListener
	{
	public:		
		ReplacementPolicy(Cache *cache)
		{
			cache->setReplacementPolicy(this);	
		}
		//Removes the victim from the 
		//Returns null if none is left in the cache.		
		virtual ICacheEntry* getVictimItem()=0;

		virtual void newData(ICacheEntry &cacheEntry)=0;

		/**
		* Returns the next candidate victim however does
		* not alter anything in the state of the replacement policy
		* \return 
		*/
		virtual ICacheEntry* peek() = 0;

		/**
		* Used for debugging purposes
		* \return 
		*/
		virtual void checkForError() = 0;

		virtual ~ReplacementPolicy(){};

		//Called for all the cache hits	
		virtual void hitCacheEntry(ICacheEntry& cacheEntry) = 0;

		//Called to clear the contents of the cache, used only in extra-ordinary circumstances
		virtual void clearCache()=0;
		
		virtual void ifRemainder(bool flag) = 0;

		/**
		* This call removes the list of cache entries
		* from the cache. Used in case there are many or clauses
		* in the remainder query.
		* Removes the entries from the LRU list and adds them to the 
		* free list.
		* \param listCacheEntry 
		*/
		virtual void removeEntries(std::list<ICacheEntry*>& listCacheEntry) = 0;

	};

}
#endif