#ifndef _STATISTICS_GATHERER_H
#define _STATISTICS_GATHERER_H

#include <ServedItemVisitor.h>
#include <ILoadListener.h>
#include <cache/prefetcher/PrefetchVisitor.h>
#include <db_GLOBAL.h> //Defines all the macros for Prefetch Strategies
#include <sstream>
#include <log/log.h>
#include <Timer/Timer.h>

namespace cache
{
	//Forward References
	class ICacheEntry;
	class Brush;
	class Prefetcher;

	class StatisticsGatherer : public ServedItemListener,
		public ILoadListener,
		public PrefetchVisitor
	{
	public:
		//Constructs a StatisticsGatherer
		StatisticsGatherer();

		//Declaring the copy constructor and the = operator
		StatisticsGatherer(StatisticsGatherer& stat);

		StatisticsGatherer& operator =(StatisticsGatherer &statisticsGatherer);


		/**
		* Implements the served Item visitor.
		* This method calculates the usage statistics
		* Mainly the misspredicted count
		* and the correctly predicted count.
		* \param cacheEntry 
		*/
		void servedItem(ICacheEntry& cacheEntry) ;
		
		/**
		* This will be called in the context of the main thread
		* Implement the IEndLoadListener
		* This is used to update the table with the usage statistics
		*/
		void stopLoad();
		
		/************************************************************************/
		/*  This starts the  timer and used to keep track of the resp time      */ 
		/*																		*/
		/************************************************************************/
		void startLoad();
		
		/**
		* This will be called in the context of the prefetch thread
		* Copy the state of the prefetcher so that it can be used 
		* to calculate the statistics at the end of the next user load
		* query
		* \param Prefetcher
		* \param *prefetcher 
		*/
		void prefetcherShuttingDown(Prefetcher *prefetcher);
		
		/**
		* Destroys the statistics gatherer
		* \return 
		*/
		virtual ~StatisticsGatherer();
			

	private:	
		//Stores the pointer to the vector that store the prefetch brushes
		std::vector<Brush*> prefetchBrushes;
		//Stores the number of nodes prefetched
		unsigned int numberOfNodesPrefetched;
		//Stores the number of correctly predicted cache entries
		unsigned int correctlyPredicted;
		//Stores the number of miss predicted cache entries
		unsigned int misPredicted;
		//Stores the number of not predicted cache entries
		unsigned int notPredicted;
		//Stores the statistics logger
		debug::log::CMessageGroup statisticsLogger;
		//Stores the Timer object
		Timer timer;
	};

}
#endif //_STATISICS_GATHERER_H
