#ifndef CHierDisplayInformation_H
#define CHierDisplayInformation_H

#include <xmdv/colormap.h>
#include <stdio.h>
#define STYLE_NONE 0x01
#define STYLE_HIGHLIGHT 0x02
#define STYLE_COLORMAP 0x04
#define STYLE_THICK 0x08
#define STYLE_DIM 0x10

#ifdef XMDV_USE_BACKEND
#include <cache/ConfigurationStore.h>
#include <cache/CacheController.h>
#include <cache/CacheBuilder.h>
#endif

class CHierDisplayInformation
{
	/*
	*	This class stores all the global information that is
	*  required about the data base. Like the, 
	a) data table
	b) contour table.  
	c) user name
	d) passwd
	e) dsn
	*  f) The reference to the backend cache structure null if not used
	*/
	class DataBaseInfo
	{
		public:

		DataBaseInfo();
		//A method used to reset the contents of the data base info.
		void reset();
		/**User trace file*/
		FILE *histfptr;
		#ifdef XMDV_USE_BACKEND
			/*This is set by the appConnectDataBase and then used by all the 
			* modules that use the cache, null if cache is not used*/
			cache::ICacheController* cacheController;
			//Store a reference to the dbConfigStore
			cache::DBConfigStore* dbConfigStore;
			//Stores a reference to the cache config store
			cache::CacheConfigStore* cacheConfigStore;
			//Store a reference to the cache builder have to delete this
			cache::CacheBuilder* cacheBuilder;
		#endif //XMDV_USE_BACKEND
		//Stores the active flag indicates whether the both the hier and the database is active
		bool activeFlag;
		//Destroys the data base info
		~DataBaseInfo();
	};	

public:
	int tree_height;
	
	bool SHOW_BANDS;
	bool SHOW_MEANS;
	bool SHOW_BRUSH_BKGRD;
	bool COLORED_BANDS;
	bool EXTENT_BRUSH;

	double brushed_radius;
	double cluster_radius;
	/**Stores the radius of the root node*/
	double root_radius;	

	double tree_brush[2];

	int normal_style;
	int highlight_style;

	double fade_factor;
	double extent_scaling_fac;
	double density;

	colormap *CMAP;  

	int dimension;

	DataBaseInfo databaseInfo;
	
	//Stores a flag indicating if the data base is active.
	bool database;

public:
	CHierDisplayInformation();
	~CHierDisplayInformation();
	
	void Reset();
	void SetBackend(bool backend);
};


#endif
