
#ifndef __CHIERNAVHELPER_H__
#define __CHIERNAVHELPER_H__

// SYSTEM INCLUDES
//
#include<math.h>
#include <iostream>

// PROJECT INCLUDES
//
#include <xmdv/CHierDisplay.h>
#include <xmdv/ARR.h>

// DEFINES
//
#define CONTOUR_RES 25

///
/// This an abstract class class that declares all the methods required to 
/// generate treecontour structure
///
///
class CHierNavHelper
{

public:    
	///
	/// Virtual Destructor
	///
	//virtual ~CHierNavHelper() = 0 {}
	virtual ~CHierNavHelper() {}

	///
	/// A function that finds the number of levels in the cluster tree
	/// 
	/// \returns number of level in the cluster tree
	/// 
	virtual int FindTreeHeight() = 0;

	///
	/// A function that evaluates the level values and populates the array
	/// contour radius
	///
	virtual void FindRadiusDistribution(ARR<double> &contour_radius)=0;

	///
	/// A function that popultates the TreeContour array.
	/// That is evaluates the complete contour and popultate the array 
	/// tree_countour
	///
	virtual void FindTreeContour(ARR< ARR<double> > &tree_contour,
								 ARR< ARR<double> > &tree_color,
								 ARR<double> contour_radius) = 0;

	///
	/// Returns the radius of the root node
	///
	/// \return radius of root node
	virtual double FindRootRadius()=0;

	//Just prints out the ARR given as argument
	//Used for debugging purposes
	template <class T>
	void debugArr(ARR<T> arr)
	{
		for(int i=0;i<arr.Len();i++)
		{
			std::cout<<arr[i]<<" ";
		}
	}
};

#endif //__CHIERNAVHELPER_H__
