////////////////////////////////////////////////////////////////// COPYRIGHT //
//
//  XmdvTool Version 4.0 - A program for Multivariate Data Display
//
//  Copyright 1994-1999, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty. 
//  In no event will the authors be held liable for any damages arising 
//  from the use of this software.
// 
//  Authors:
//      Matthew O. Ward, Allen R. Martin, Ying-Huey Fua
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~xmdv
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to the
//  XmdvTool Development Team: <matt@cs.wpi.edu>.
//
////////////////////////////////////////////////////////////////// COPYRIGHT //

#ifndef VEC2_H
#define VEC2_H

#include <math.h>
#include <iostream>

class Vec2
{
    public:
        Vec2(double x=0, double y=0) :
            X(x), Y(y)
        {}

        double operator[](int i) const { return ((double*)&X)[i]; }
        double &operator[](int i) { return ((double*)&X)[i]; }
        friend Vec2 operator-(const Vec2 &v1, const Vec2 &v2); 
        friend Vec2 operator+(const Vec2 &v1, const Vec2 &v2); 
        friend double operator*( const Vec2 & v1, const Vec2& v2);
        friend Vec2 operator*( double x, const Vec2& v2);
        
        friend std::ostream &operator<<(std::ostream &out, const Vec2& v);
                
        double norm() const;
        void normalize();
        void Set(double x,double y) { X=x; Y=y; }
        
    public:
        double X,Y;
};

inline Vec2 operator- (const Vec2 &v1, const Vec2 &v2 )
{
    return Vec2(v1.X - v2.X, v1.Y - v2.Y );
}

inline Vec2 operator+ (const Vec2 &v1, const Vec2 &v2 )
{
    return Vec2(v1.X + v2.X, v1.Y + v2.Y );
}

inline std::ostream &operator<<(std::ostream &out, const Vec2 &v)
{
    out << " [ " << v.X << " " << v.Y << " ] ";
    return out;
}

inline double Vec2::norm() const 
{
    return sqrt(X*X + Y*Y);
}

inline void Vec2::normalize()
{
    double mag = norm();
    X /= mag;
    Y /= mag;
}

inline double operator*( const Vec2 &v1, const Vec2 &v2)
{
    return v1.X*v2.X + v1.Y*v2.Y;
}

inline Vec2 operator*( double x, const Vec2 &v2)
{
    return Vec2( x*v2.X, x*v2.Y );
}

#endif
