//modified by Jing Yang after 7/6/2001
#include <math.h>
#include <string.h>
#ifdef _WIN32
#include <windows.h>
#else
#include <unistd.h>
#endif

#include <xmdv/CDimReducer.h>
#include <xmdv/XmdvTool.h>
#include <xmdv/OnlineDimClustering.h>

CDimReducer::CDimReducer()
{
    data_array = NULL;
    entries_array = NULL;
	dim_tree = NULL;
	representative_dim_list = NULL;

    data_strategy = USE_CLUSTERS;
    tree_strategy = REVERSE_DATACLUSTERING;
    axis_color_strategy = UNIFORM_COLOR;
    status = NORMAL;

	extent_scaling_fac=.3;
	density = 0.5;
}


CDimReducer::~CDimReducer()
{
    if ( dim_tree != NULL )
        delete dim_tree;
    g_globals.dtree = NULL;
    if ( representative_dim_list != NULL )
        delete representative_dim_list;

    if ( data_array != NULL )
    {
        for ( int i = 0; i < data_num; i++ )
        {
            delete [] data_array[i];
        }
        delete [] data_array;
    }

    if ( entries_array != NULL )
    {
        for ( int i = 0; i < data_num; i++ )
        {
            delete [] entries_array[i];
        }
        delete [] entries_array ;
    }
	hierDeleteClusterTree();
}


bool CDimReducer::GetData()
{
    if ( data_array != NULL )
    {
        for ( int i = 0; i < data_num; i++ )
        {
            delete [] data_array[i];
        }
        delete [] data_array;
    }
    if ( entries_array != NULL )
    {
        for ( int i = 0; i < data_num; i++ )
        {
            delete [] entries_array[i];
        }
        delete [] entries_array ;
    }

    switch (data_strategy) 
    {
        case USE_CLUSTERS:
        default: 
            return GetDataFromClusters();
    }
}


void CDimReducer::CollectClustersOnContour(double contour_radius, const Cluster &node, CLinkList* data_linkList, CLinkList* entries_linkList)
{
    int i;
    if ( node.num_children == 0 || node.sqrt_radius < contour_radius )
    {
	    double* data = new double [g_globals.dataset->dims];
        int *entries = new int[1];

        ARR<double> mapped_data;
        mapped_data.SetLen(g_globals.dataset->dims);

        for ( i=0; i < g_globals.dataset->dims; i++)
        {   
            g_globals.dataset->MapData(node.mean, mapped_data);
			data[i] = mapped_data[i];
        }

        entries[0] = node.entries;
        data_linkList->AddNodeAtEnd(data);
        entries_linkList->AddNodeAtEnd(entries);
        return;
    }
    else
    {
        for ( i = 0; i < node.num_children; i++ )
        {
            CollectClustersOnContour(contour_radius, g_globals.btree->nodes[node.child[i]], data_linkList, entries_linkList);
        }
        return;
    }
}


bool CDimReducer::GetDataFromClusters()
{
    double contour_radius = 0.02;
    CLinkList* data_linkList = new CLinkList();
    CLinkList* entries_linkList = new CLinkList();
 /*   if ( hier_info->database )
    {
 	    prepare( 0, 1, contour_radius, NO_P_ENTRIES-1);

	    ENTRY *entry;
	    for (init(); (entry=get_one())!=NULL; )
	    {
		    double* data = new double [g_globals.dataset->all_dims];
            for (int i=0; i<g_globals.dataset->all_dims; i++)
            {
			    data[i] = entry->data->D[i];
            }
            data_linkList->AddNodeAtEnd(data);
	    }
    }
    else
  */  {
        if ( ! g_globals.dataset )
            return false;
        
        hierInitClusterTree();
        if ( !g_globals.btree )
            return false;
        CollectClustersOnContour(contour_radius, g_globals.btree->nodes[0], data_linkList, entries_linkList);
    }

    data_array = (double**) data_linkList->TransToArray();
    entries_array = (int**) entries_linkList->TransToArray();
    data_num = data_linkList->GetLength();
    delete data_linkList;
    delete entries_linkList;
    return true;
}


void CDimReducer::RefineData()
{
    if ( data_num <= 0 )
        return;

    double *min, *max, *diff;

    min = new double [g_globals.dataset->dims];
    max = new double [g_globals.dataset->dims];
    diff = new double [g_globals.dataset->dims];

    int i, j;
    for ( j = 0; j < g_globals.dataset->dims; j++ )
    {
        min[j] = max[j] = data_array[0][j];
        for ( i = 1; i < data_num; i++ )
        {
            if ( min[j] > data_array[i][j] )
                min[j] = data_array[i][j];
            else if ( max[j] < data_array[i][j] )
                max[j] = data_array[i][j];
        }

        if ( fabs(max[j]-min[j]) < 1e-5 )
            min[j] = max[j] - 0.05;
        diff[j] = 1/(max[j]-min[j]);

        for ( i = 0; i < data_num; i++ )
        {
            data_array[i][j] = (data_array[i][j]-min[j])*diff[j];
        }
    }

    data_min = new double [data_num];
    data_max = new double [data_num];
    
    for ( i = 0; i < data_num; i++ )
    {
        data_min[i] = data_max[i] = data_array[i][0];
        for ( j = 1; j < g_globals.dataset->dims; j++ )
        {
            if ( data_min[i] > data_array[i][j] )
                data_min[i] = data_array[i][j];
            else if ( data_max[i] < data_array[i][j] )
                data_max[i] = data_array[i][j];
        }
    }
    
}


void CDimReducer::SaveDataToFile(const char * fname)
{
    int i, j;

    FILE *fptr = fopen(fname,"w");
    fprintf(fptr, "%d %d\n", data_num, g_globals.dataset->dims);

    for ( i = 0; i < data_num; i++ )
    {
        fprintf(fptr, "%lf %lf\n", data_min[i], data_max[i]);
    }

    delete [] data_min;
    delete [] data_max;
    
    for ( j = 0; j < g_globals.dataset->dims; j++ )
    {
        fprintf(fptr, "%d ", j);
        for ( i = 0; i < data_num; i++ )
        {
            fprintf(fptr, "%lf ", data_array[i][j]);
        }
        fprintf(fptr, "\n");
    }

    for ( i = 0; i < data_num; i++ )
    {
        fprintf(fptr, "%d ", entries_array[i][0]);
    }
    fprintf(fptr, "\n");
    
    fclose(fptr);
}

bool CDimReducer::BuildDimClusterTree()
{
    int N = strlen(g_globals.dataset->datafile);
	char *dcfile;
    dcfile = new char[N+1];
    strcpy(dcfile, g_globals.dataset->datafile);
    strcpy(&dcfile[N-4],".dc");
    FILE *fp;
    if ( ( fp = fopen(dcfile, "r" ) ) != NULL )
    {
        fclose(fp);
    }
    else
    {
        if ( !GetData() )
            return false;
        RefineData();
        char* temp_file = "temp__";
        SaveDataToFile(temp_file);
        bool succ = OnlineClusterDimension(temp_file, dcfile, 30, int(g_globals.dataset->data_size*0.7 ));
#ifdef _WIN32
        DeleteFile(temp_file);
#else
        unlink(temp_file);
#endif
        if ( !succ )
            return false;
    }

    if ( dim_tree != NULL )
        delete dim_tree;
	dim_tree = new CInterRingDimClusterTree();
	if ( dim_tree->BuildInterRingDimClusterTreeFromFile(dcfile) == false )
    {
        delete dim_tree;
        return false;
    }
    if ( representative_dim_list != NULL )
        delete representative_dim_list;
    representative_dim_list = new CLinkList;

	delete [] dcfile;
    return true;
}

CInterRingDimClusterTree* CDimReducer::GetDimClusterTree()
{
    if ( !BuildDimClusterTree() )
        return NULL;

    return dim_tree;
}

void CDimReducer::AddRepresentativeDim(CInterRingDimCluster* current_node)
{
    if ( dim_tree == NULL )
        return;
    representative_dim_list->AddNodeAtEnd(current_node);    
}


void CDimReducer::DeleteRepresentativeDim(CInterRingDimCluster* current_cluster)
{
    if ( dim_tree == NULL )
        return;
	representative_dim_list->DeleteElement(current_cluster);
}


int CDimReducer::SetStatus(const char *type)
{
    if (strcmp(type,"normal")==0)
    {
        status = NORMAL;
        g_globals.dataset->CountDims();
        g_globals.dataset->RemapAllData();
    }
/*    else if (strcmp(type,"distort")==0)
        screen_distortion->distortion_status = true; 
    else if (strcmp(type,"rotate")==0)
        screen_distortion->distortion_status = true; 
    else if (strcmp(type,"modify")==0)  
    {
        SetStatus("normal");
        status = MODIFY;
        screen_distortion->distortion_status = false; 
    }
    else if (strcmp(type,"select")==0)
    {
//        screen_distortion->distortion_status = false; 
        SetStatus("normal");
        status = SELECT;
    }
    else if (strcmp(type,"customize")==0)
    {
        screen_distortion->distortion_status = false; 
    } */
    else if (strcmp(type,"apply")==0)
    {
        if ( status == APPLY )
            return -1;
        if ( representative_dim_list->GetLength() <= 0 )
            return -2;
        if ( representative_dim_list->GetLength() > g_globals.dataset->all_dims )
            return -3;
        SaveMinMax();
        status = APPLY;
        g_globals.dataset->dims = representative_dim_list->GetLength();
        g_globals.dataset->RemapAllData();
        g_globals.dim_reducer->ResetAxisColor();
        return 1;
    }
    return 0;
}


void CDimReducer::SaveMinMax()
{
    dim_max.SetLen(g_globals.dataset->all_dims);
    dim_min.SetLen(g_globals.dataset->all_dims);
    g_globals.dataset->MapData(g_globals.dataset->all_dim_max, dim_max);
    g_globals.dataset->MapData(g_globals.dataset->all_dim_min, dim_min);
}

// PAUL: lang_opt
// Added const &'s so that the copy constructor isn't called
void CDimReducer::MapData( const ARR<double> & all_data, ARR<double> &data)
{
    int i = 0;
    CLinkListNode *node = representative_dim_list->GetFirst();
    while ( node != NULL )
    {
        data[i] = ((CInterRingDimCluster *)(node->element))->Calculate(all_data);
        node = representative_dim_list->GetNext(node);
        i++;
    }
}

// PAUL: brush_org
// Map a brush using CDimReducer in an analogous fashion
void CDimReducer::MapData(const ARR<int> & index, const Brush * all_brush, Brush * brush)
{
	static ARR<double> all_pos, all_size, all_offset;
	const int len = index.Len();
	all_pos.SetLen(len);
	all_size.SetLen(len);
	all_offset.SetLen(len);
	int i;
	for (i = 0; i < len; ++i)
	{
		int idx = index[i];
		all_pos[i] = all_brush->Pos(idx);
		all_size[i] = all_brush->Size(idx);
		all_offset[i] = all_brush->BoundOffset(idx);
	}	

    i = 0;
    CLinkListNode *node = representative_dim_list->GetFirst();
	CInterRingDimCluster * pIRDC;
    while ( node != NULL )
    {
		pIRDC = (CInterRingDimCluster *)node->element;
        brush->SetTriple(i, pIRDC->Calculate(all_pos),
							pIRDC->Calculate(all_pos),
							pIRDC->Calculate(all_pos));

        node = representative_dim_list->GetNext(node);
        i++;
    }
}

void CDimReducer::MapDataWithDimClusterExtends( const ARR<double> & all_data, ARR<double> &mean, 
                                                ARR<double> &max, ARR<double> &min)
{
    int i = 0;
    CLinkListNode *node = representative_dim_list->GetFirst();
    while ( node != NULL )
    {
        ((CInterRingDimCluster *)(node->element))->CalculateWithDimClusterExtends(all_data, mean[i], max[i], min[i]);
        node = representative_dim_list->GetNext(node);
        i++;
    }
}


void CDimReducer::MapData(const ARR<int> & all_data, ARR<int> &data)
{
    int i = 0;
    CLinkListNode *node = representative_dim_list->GetFirst();
    while ( node != NULL )
    {
        data[i] = ((CInterRingDimCluster *)(node->element))->Calculate(all_data);
        node = representative_dim_list->GetNext(node);
        i++;
    }
}



void CDimReducer::MapData( const ARR<NominalField*> & all_nominal_fields, ARR<NominalField*> &nominal_fields)
{
    int i = 0, okc_id;
    CLinkListNode *node = representative_dim_list->GetFirst();
    while ( node != NULL )
    {
        okc_id = ((CInterRingDimCluster *)(node->element))->okc_id;
        if ( okc_id >= 0 ) 
            nominal_fields[i] = all_nominal_fields[okc_id];
        else 
            nominal_fields[i] = 0;
        node = representative_dim_list->GetNext(node);
        i++;
    }
}


void CDimReducer::MapData( const ARR<char *> & all_names, ARR<char *> &names)
{
    int i = 0;
    CLinkListNode *node = representative_dim_list->GetFirst();
    while ( node != NULL )
    {
        names[i] = ((CInterRingDimCluster *)(node->element))->name;
        node = representative_dim_list->GetNext(node);
        i++;
    }
}


void CDimReducer::ExtendNodeInRepresentativeDimList(int N)
{
    if ( status != APPLY )
        return;

    CInterRingDimCluster *cluster;
    CLinkListNode *node;
    cluster = (CInterRingDimCluster *)(representative_dim_list->GetElement(N));
    node = representative_dim_list->Get(N);
    if ( cluster->okc_id >= 0 || cluster->children == NULL )
        return;

    CInterRingDimCluster *child = (CInterRingDimCluster *)cluster->children;
    while ( child != NULL )
    {
        representative_dim_list->DeleteElement(child);
        child->selected = true;
        representative_dim_list->AddNodeAfter(child, node);
        child = (CInterRingDimCluster *)child->siblings;
    }

    representative_dim_list->DeleteNode(node);
    cluster->selected = false;

    g_globals.dataset->dims = representative_dim_list->GetLength();
    g_globals.dataset->RemapAllData();
}

void CDimReducer::ShinkNodeInRepresentativeDimList(int N)
{
    if ( status != APPLY )
        return;

    CInterRingDimCluster *cluster;
    CLinkListNode *node;
    cluster = (CInterRingDimCluster *)(representative_dim_list->GetElement(N));
    node = representative_dim_list->Get(N);
    if ( cluster->parent == NULL )
        return;

    CInterRingDimCluster *parent = (CInterRingDimCluster *)cluster->parent;
    parent->selected = true;
    representative_dim_list->AddNodeAfter(parent, node);

    CInterRingDimCluster *child = (CInterRingDimCluster *)parent->children;
    while ( child != NULL )
    {
        representative_dim_list->DeleteElement(child);
        child->selected = false;
        child = (CInterRingDimCluster *)child->siblings;
    }

    g_globals.dataset->dims = representative_dim_list->GetLength();
    g_globals.dataset->RemapAllData();
}

void CDimReducer::ResetRepresentativeDimList()
{   
    CInterRingDimCluster *cluster;
    int n = representative_dim_list->GetLength();
    for ( int i = n-1; i >= 0; i-- )
    {
        cluster = (CInterRingDimCluster *)(representative_dim_list->GetElement(i));
        cluster->selected = false;
        representative_dim_list->Delete(i);
    }
}

double CDimReducer::GetDimDisagreement(int N)
{
    if ( N >= representative_dim_list->GetLength() )
        return 0;

    CInterRingDimCluster *cluster;
    cluster = (CInterRingDimCluster *)(representative_dim_list->GetElement(N));

    return cluster->GetRadius();
}

int CDimReducer::GetEntries(int N)
{
    if ( N >= representative_dim_list->GetLength() )
        return 0;

    CInterRingDimCluster *cluster;
    cluster = (CInterRingDimCluster *)(representative_dim_list->GetElement(N));

    return cluster->entries;
}

RGBt CDimReducer::GetClusterColor(int N)
{
    if ( N >= representative_dim_list->GetLength() )
        return 0;

    CInterRingDimCluster *cluster;
    cluster = (CInterRingDimCluster *)(representative_dim_list->GetElement(N));

    return cluster->color;
}

bool CDimReducer::IsLeafDim(int N)
{
    if ( N >= representative_dim_list->GetLength() )
        return true;

    CInterRingDimCluster *cluster;
    cluster = (CInterRingDimCluster *)(representative_dim_list->GetElement(N));
    if ( cluster->okc_id < 0 )
        return false;
    else 
        return true;
}


void CDimReducer::SizeStructureSelectNode(int threshold, bool mark)
{
    current_node->ResetSelected( !mark );
    current_node->SizeStructureSelect(threshold, mark);
}

/*
void CDimReducer::RadiusStructureSelectNode(double threshold, bool mark)
{
    current_node->ResetSelected( !mark );
    current_node->RadiusStructureSelect(threshold, mark);
}
*/

void CDimReducer::SetAxisColorStrategy(const char* arg)
{
    if ( strcmp(arg, "uniform_color") == 0 )
        axis_color_strategy = UNIFORM_COLOR;
    else 
        axis_color_strategy = NODE_COLOR;
}


void CDimReducer::ResetAxisColor()
{
    if ( axis_color_strategy == UNIFORM_COLOR )
    {
        for ( int i = 0; i < g_globals.dataset->dims; i++ )
        {
            g_globals.dataset->axis_color[i] = interface_color[GRID1];
        }
    }
    else
    {
        for ( int i = 0; i < g_globals.dataset->dims; i++ )
        {
            g_globals.dataset->axis_color[i] = GetClusterColor(i);
        }
    }
}

void CDimReducer::GetSpacingDistances( double *distances )
{
    int i;
    int length = representative_dim_list->GetLength();

    CInterRingDimCluster *node1, *node2;
    node1 = (CInterRingDimCluster *)representative_dim_list->GetElement(0);
    for ( i = 0; i < length-1; i++ )
    {
        node2 = (CInterRingDimCluster *)representative_dim_list->GetElement(i+1);
        distances[i] = dim_tree->GetDistance(node1, node2);
        node1 = node2;
    }
    node2 = (CInterRingDimCluster *)representative_dim_list->GetElement(0);
    distances[i] = dim_tree->GetDistance(node1, node2);
}


