
#include <string.h>

#include <xmdv/CFlatDimstack.h>
#include <xmdv/ogl.h>
#include <xmdv/XmdvTool.h>
#include <xmdv/brush.h>
#include <xmdv/brush_oper.h>
#include <xmdv/multi_select.h>
#include <xmdv/okcdata.h>


CFlatDimstack::CFlatDimstack(CFlatDisplayInformation* flat_info) : CFlatDisplay(flat_info)
{
}

CFlatDimstack::~CFlatDimstack()
{
}

//
// Determine if this value (val) for this dimension (dim) is covered by
// the brush.
//
static bool is_within(int val, int dim, int brush)
{
    double fval = val;
    double range = g_globals.dataset->dim_max[dim] - g_globals.dataset->dim_min[dim];
    double b_pos = g_globals.dataset->cardinality[dim] * (g_globals.brushes[brush].Pos(dim)-g_globals.dataset->dim_min[dim]) / range;
    double b_size = .5 * g_globals.dataset->cardinality[dim] * g_globals.brushes[brush].Size(dim) / range;
    double start = b_pos - b_size;
    double end = b_pos + b_size;
    if (end > fval && end <= (fval+1.0)) return true;
    if (start >= fval && start < (fval+1.0)) return true;
    if (start < fval && end > (fval+1.0)) return true;
    return false;
}

//
// Recursive check for brush coverage, 2 dimensions at a time
//
void CFlatDimstack::rec_fill(grCanvas *canvas, double basex, double basey, int nextx, int nexty,
          ARR<double> &size, int brush)
{
	// If we had an even number of dimensions and are at the bottom of the
    // recursion, draw a rectangle in that data spot, as it is covered.
    //
    if (nextx == g_globals.dataset->dims)
    {
        canvas->FillRectangle(basex, g_globals.screen_distortion->dimstack_Y_size - basey - g_globals.screen_distortion->dimstack_Y_blocksize, 
							  g_globals.screen_distortion->dimstack_X_blocksize, g_globals.screen_distortion->dimstack_Y_blocksize);
        return;
    }

	// PAUL: draw_org
	// If we have recursed enough, just fill the space rather than go down in near-endless recursion
	if (nextx >= sensible_limit)
	{
		for (int i=0; i<g_globals.dataset->cardinality[nextx]; i++)
		{
            if (is_within(i, nextx, brush))
				canvas->FillRectangle(basex, g_globals.screen_distortion->dimstack_Y_size - basey - sensible_Y_blocksize, 
									  sensible_X_blocksize, sensible_Y_blocksize);
		}
		return;
	}

    // If we had an odd number of dimensions and are at the bottom of the
    // recursion, check each descrete value for brush overlap and draw rectangles
    // over covered spots.
    //
    int i;
    if (nexty == g_globals.dataset->dims)
    {
        for (i=0; i<g_globals.dataset->cardinality[nextx]; i++)
		{
            if (is_within(i, nextx, brush))
                canvas->FillRectangle(basex + i * size[nextx],
                    g_globals.screen_distortion->dimstack_Y_size - basey - g_globals.screen_distortion->dimstack_Y_blocksize, g_globals.screen_distortion->dimstack_X_blocksize, 
					g_globals.screen_distortion->dimstack_Y_blocksize);
        }
        return;
    }

    // Continue recursion for each bucket of the current 2 dimensions
    //
    for (i=0; i<g_globals.dataset->cardinality[nextx]; i++)
		// PAUL: draw_org
		// Move an is_within() call to the outside
		if (is_within(i, nextx, brush))	// Moved from the inner loop
			for (int j=0; j<g_globals.dataset->cardinality[nexty]; j++)
				if (is_within(j, nexty, brush))
					rec_fill(canvas, basex + i * size[nextx], basey + j * size[nexty], nextx+2, nexty+2, size, brush);
}

//
// Display brush coverage by descretizing brush in each dimension and check
// where in range of that dimension it overlaps.  Roundoff error in the
// discretization process can reduce the visual effect of brush coverage.
//
void CFlatDimstack::DrawSingleBrush(Brush *brush)
{
	// PAUL: draw_org
	// Move BeginDraw to the outside, so it is not called every time a rectangle is drawn
    canvas->SetForeground( interface_color[BRUSH1+brush->num]);
	canvas->BeginDraw();
	// For the outermost dimensions...
    for (int i=0; i<g_globals.dataset->cardinality[0]; i++) 
    {
		// PAUL: draw_org
		// Move one of the inner is_withins to the outside
		if (is_within(i, 0, brush->num) == 1)
		{
			for (int j=0; j<g_globals.dataset->cardinality[1]; j++)
			{
				// if the brush intersects those dimensions recursively
				// call to fill
				if (/*(is_within(i, 0, brush->num) == 1) &&*/ (is_within(j, 1, brush->num) == 1))
					rec_fill(canvas, i * g_globals.screen_distortion->dimstack_size[0], j * g_globals.screen_distortion->dimstack_size[1], 
							2, 3, g_globals.screen_distortion->dimstack_size, brush->num);
			}
		}
    }
	canvas->EndDraw();
}

//
// Compute the descretized value from an original data value for a particular
// dimension.  Display will be messed up if it is out of bounds.  You then
// should fix the range for that dimension in your data file.
//
int get_contrib(int ind, double data)
{
    int res = int( ((data - g_globals.dataset->dim_min[ind]) / (g_globals.dataset->dim_max[ind] - g_globals.dataset->dim_min[ind])) *
              (double)(g_globals.dataset->cardinality[ind]));
/*
    if (res >= g_globals.dataset->cardinality[ind] || res < 0) {
        printf("Data point out of bounds: dimension %d: [%lf, %lf]: %lf\n", ind,
            g_globals.dataset->dim_min[ind], g_globals.dataset->dim_max[ind], data);
    }
*/
    return res;
}

void CFlatDimstack::DrawSingleDimstackBin(grCanvas *canvas, const ARR<double> &data, const ARR<double> &size)
{
	// PAUL: lang_opt
	// Make contrib static, so that it is not reallocated every time the function
	// is called for a single drawing event
    static ARR<int> contrib;
	contrib.SetLen(g_globals.dataset->dims);
    int j;
    for (j=0; j<g_globals.dataset->dims; j++)
        contrib[j] = get_contrib(j, data[j]);

    double endx = 0, endy = 0;
    for (j=0; j<g_globals.dataset->dims; j++)
    {
        endx += contrib[j] * size[j];
        j++;
        if (j < g_globals.dataset->dims)
            endy += contrib[j] * size[j];
    }
	//printf("Fill(%lf, %lf)\n", endx,  g_globals.screen_distortion->dimstack_Y_size - endy - g_globals.screen_distortion->dimstack_Y_blocksize);
    canvas->FillRectangle(endx, g_globals.screen_distortion->dimstack_Y_size - endy - g_globals.screen_distortion->dimstack_Y_blocksize, 
						  g_globals.screen_distortion->dimstack_X_blocksize, g_globals.screen_distortion->dimstack_Y_blocksize);
}

void CFlatDimstack::DrawSingleData(ARR<double> REF data, int datano)
{
    DrawSingleDimstackBin(canvas, data, g_globals.screen_distortion->dimstack_size);
}



void CFlatDimstack::Draw()
{
    if (!g_globals.dataset) return;

    if (g_globals.dataset->dims<2) {
        return;
    }

    CFlatDisplay::Draw();
}

void CFlatDimstack::SetParameters()
{
    CFlatDisplay::SetParameters();

	// PAUL: draw_org
	// Calculate values for sensible drawing of dimstack
	// When the number of horizontal divisions reaches a certain
	// point, the drawing ceases to limit the exponential nature
	// of drawing a dimstack
	int divisions = 1;
	sensible_X_blocksize = sensible_Y_blocksize = 1;

	sensible_limit = g_globals.dataset->dims - 1;
	int i = 0;
	for (; i < g_globals.dataset->dims; i += 2)
	{
		divisions *= g_globals.dataset->cardinality[i];
		// PAUL: change this value to allow deeper dimstack drawing
		if (divisions >= 256)
		{
			sensible_limit = i;
			break;
		}
		// These values are needed for properly drawing the brush
		// at a certain depth
		sensible_X_blocksize *= g_globals.dataset->cardinality[i];
		if (g_globals.dataset->cardinality.Len() > i + 1)
			sensible_Y_blocksize *= g_globals.dataset->cardinality[i+1];
	}

	// PAUL: draw_org
	// Is the number of dimensions odd?
	if (i > 2 && i >= g_globals.dataset->dims && g_globals.dataset->dims % 2)
	{
		sensible_X_blocksize /= g_globals.dataset->cardinality[i-2];
	}

	// blocksize is a fraction out of 1
	sensible_X_blocksize = 1/ sensible_X_blocksize;
	sensible_Y_blocksize = 1/ sensible_Y_blocksize;
}

void CFlatDimstack::DrawAxes()
{
    DrawDimstackAxes(canvas);
}

void CFlatDimstack::DrawDimstackAxes(grCanvas *canvas)
{
	canvas->bypassAll(true);

    glLineWidth(1.);
    glShadeModel(GL_SMOOTH);
    // Compute the size of the blocks needed for the slowest dimension
    //
    double x_size = g_globals.screen_distortion->dimstack_X_size;
    double y_size = g_globals.screen_distortion->dimstack_Y_size;

    // Draw up to 3 levels of grid marks, based on the number of dimensions
    if (g_globals.dataset->dims >= 6)
    {
        canvas->SetForeground(interface_color[GRID3]);
		// PAUL: draw_org
		// Move begindraw to the outside
		canvas->BeginDraw();
        for (double x = 0;x <= x_size; x = x + g_globals.screen_distortion->dimstack_size[4])
            canvas->DrawLine((float)x, 0.0f, (float)x, (float)y_size);

        for (double y = 0;y <= y_size; y = y + g_globals.screen_distortion->dimstack_size[5])
            canvas->DrawLine(0.0f, (float)y, (float)x_size, (float)y);
		canvas->EndDraw();
    }

    if (g_globals.dataset->dims >= 4)
    {
        canvas->SetForeground(interface_color[GRID2]);
		// PAUL: draw_org
		// Move begindraw to the outside
		canvas->BeginDraw();
        for (double x = 0;x <= x_size; x = x + g_globals.screen_distortion->dimstack_size[2])
            canvas->DrawLine((float)x, 0.0f, (float)x, (float)y_size);

        for (double y = 0;y <= y_size; y = y + g_globals.screen_distortion->dimstack_size[3])
            canvas->DrawLine(0.0f, (float)y, (float)x_size, (float)y);
		canvas->EndDraw();
    }

    canvas->SetForeground(interface_color[GRID1]);

	// PAUL: draw_org
	// Move begindraw to the outside
	canvas->BeginDraw();
    for (double x = 0;x <= x_size; x = x + g_globals.screen_distortion->dimstack_size[0])
        canvas->DrawLine((float)x, 0.0f, (float)x, (float)y_size);

    for (double y = 0;y <= y_size; y = y + g_globals.screen_distortion->dimstack_size[1])
        canvas->DrawLine(0.0f, (float)y, (float)x_size, (float)y);
	canvas->EndDraw();

	glShadeModel(GL_FLAT);
	canvas->Flush();

	canvas->bypassAll(true);
}


static bool DimstackMapScreenToData(double x,double y, ARR<double> &center)
{
    // Now find the bin that the user clicked in
    int i;
    for (i=0; i<g_globals.dataset->dims; i++)
    {
        // If odd, it's in the y direction
        if (i%2)
        {
            int bin = int(y/g_globals.screen_distortion->dimstack_size[i]);

            // invert y
            int orig_bin = bin;
            bin = g_globals.dataset->cardinality[i] - 1 - bin;

            // check bounds
            if (bin < 0 || bin >= g_globals.dataset->cardinality[i])
                return false;

            // Increment data coords based on this
            center[i] = bin * (g_globals.dataset->dim_max[i] - g_globals.dataset->dim_min[i])/g_globals.dataset->cardinality[i];
            center[i] += g_globals.dataset->dim_min[i];

            // use original bin # for y decrement
            y -= orig_bin*g_globals.screen_distortion->dimstack_size[i];
        }

        // If even, it's in the x direction
        else
        {
            int bin = int(x/g_globals.screen_distortion->dimstack_size[i]);

            // check bounds
            if (bin < 0 || bin >= g_globals.dataset->cardinality[i])
                return false;

            // Increment data coords based on this
            center[i] = bin * (g_globals.dataset->dim_max[i] - g_globals.dataset->dim_min[i])/g_globals.dataset->cardinality[i];
            center[i] += g_globals.dataset->dim_min[i];

            x -= bin*g_globals.screen_distortion->dimstack_size[i];
        }
    }
    return true;
}

static void InputDimstackBrushSelect(grCanvas *canvas, int sx, int sy)
{
    Vec2 p = canvas->MapScreenToData( Vec2(sx,sy) );

    // If brushing is not on or not displayed, forget it
    if (g_globals.curr_brush->display == false || g_globals.curr_brush->enable == false)
        return;

    // Now find the bin that the user clicked in
    ARR<double> new_center(g_globals.dataset->dims);
    if ( ! DimstackMapScreenToData(p.X,p.Y,new_center) )
        return; 

    // Recenter the brush
    int i;
    for (i=0; i<g_globals.dataset->dims; i++)
        g_globals.curr_brush->SetPos(i, new_center[i]);

    // Check if any of the brush values went out of range
    for (i=0; i<g_globals.dataset->dims; i++)
    {
        // don't resize brush, but just move it so it stays in range
        if (g_globals.curr_brush->Max(i) > g_globals.dataset->dim_max[i])
            g_globals.curr_brush->SetPos(i, g_globals.dataset->dim_max[i] - g_globals.curr_brush->Size(i)/2.);
        if (g_globals.curr_brush->Min(i) < g_globals.dataset->dim_min[i])
            g_globals.curr_brush->SetPos(i, g_globals.dataset->dim_min[i] + g_globals.curr_brush->Size(i)/2.);
    }
}

static void InputDimstackMultiSelect(grCanvas *canvas, int sx, int sy)
{
    if (!multi_select) return;

    Vec2 p = canvas->MapScreenToData( Vec2(sx,sy) );

    // Now find the bin that the user clicked in
    ARR<double> center(g_globals.dataset->dims);
    if ( ! DimstackMapScreenToData(p.X,p.Y,center) )
        return;

/*
    Vec2 screen_rad = canvas->MapScreenToData( Vec2(multisel_radius,multisel_radius) ) -
                      canvas->MapScreenToData( Vec2(0,0) );
*/
    const double radN = .8;

    ARR<double> data(g_globals.dataset->dims);

    // Find all points within circle distance of mouse
    //
    for (int i=0; i<g_globals.dataset->data_size; i++)
    {
        g_globals.dataset->GetData(data, i);

        double dist = 0;
        for (int j=0;j<g_globals.dataset->dims;j++) {
            double d = (center[j] - data[j]);
            dist += d*d;
        }
        dist = sqrt(dist);

        if (dist < radN)
        {
            multiSelectAddPoint(i);

            glDrawBuffer(GL_FRONT);
            canvas->SetForeground(interface_color[PAINT]);
			CFlatDimstack::DrawSingleDimstackBin(canvas,data,g_globals.screen_distortion->dimstack_size);
            glDrawBuffer(GL_BACK);
        }
    }
}

static void DimstackCaption(grCanvas *canvas, int x, int y)
{
    Vec2 p = canvas->MapScreenToData( Vec2(x,y) );

    // Now find the bin that the user clicked in
    ARR<double> center(g_globals.dataset->dims);
    if ( ! DimstackMapScreenToData(p.X,p.Y,center) )
        return;;

    char caption[50000]="Bin center: (";

    for (int i=0; i<g_globals.dataset->dims; i++) {
        char str[100];
		//changed by Zaixian to enable the nominal field on status bar
		//the original code is: sprintf(str, "%s: %.2lf ", g_globals.dataset->names[i], center[i])
        sprintf(str, "%s: %s ", g_globals.dataset->names[i], 
			g_globals.dataset->MapValueToString(i, center[i]));
        strcat(caption, str);
    }
    strcat(caption,")");

    appShowCanvasMessage(Togl_Interp(canvas->GetWidget()),caption, canvas->name);
}

void InputDimstackButtonPress(grCanvas *canvas, int button, int x, int y)
{
    if (!g_globals.dataset) return;

    if (canvas->shiftkey && button == 0) {
        InputDimstackMultiSelect(canvas,x,y);
        return;
    } else

    if (button==2) {
        canvas->SetHandCursor();
        return;
    } else

    if (button == 0) {
        InputDimstackBrushSelect(canvas,x,y);
        BrushOperMarkDataset();
        appRefreshAllCanvas(Togl_Interp(canvas->GetWidget()));
    }
}

void CFlatDimstack::InputDimstackMouseMotion(grCanvas *canvas, int x, int y)
{
    if (!g_globals.dataset) return;

    DimstackCaption(canvas, x, y);

    if (canvas->shiftkey && canvas->button[0] == true) {
        InputDimstackMultiSelect(canvas,x,y);
        return;
    } else

    if (canvas->button[2] == true) {
        appCanvasMousePan(canvas);
    }
}

void InputDimstackButtonRelease(grCanvas *canvas, int button, int x, int y)
{
    if (canvas->shiftkey && button == 0) {
        return;
    }
    canvas->SetDefaultCursor();
}


void CFlatDimstack::InputMouseMotion(int x, int y)
{
    InputDimstackMouseMotion(canvas,x,y);
}


void CFlatDimstack::InputButtonPress(int button, int x, int y)
{
    InputDimstackButtonPress(canvas, button, x,y);
}


void CFlatDimstack::InputButtonRelease(int button, int x, int y)
{
    InputDimstackButtonRelease(canvas, button, x, y);
}

void CFlatDimstack::ClutterBasedOrder()	// added by debbie 9/2003
{
	g_globals.dataset->Clutter_Based_Order("dimstck"); // optional is neighbor definition		
}
