
#include <string.h>
#include <xmdv/CHierDisplayInformation.h>
#include <xmdv/XmdvTool.h>
#include <xmdv/colorramp.h>
#include <xmdv/okcdata.h>



CHierDisplayInformation::CHierDisplayInformation()
{
	Reset();
}

CHierDisplayInformation::~CHierDisplayInformation()
{
	delete CMAP;
}

void CHierDisplayInformation::Reset()
{
	cluster_radius = 0.5;
	brushed_radius= 0.5;
	fade_factor=0; // [-1,+1]
	extent_scaling_fac=1.0;
	density = 0.5;	
	SHOW_BANDS = false;//show_bands
	SHOW_MEANS = true;
	SHOW_BRUSH_BKGRD = true;
	COLORED_BANDS = true;

	normal_style = STYLE_COLORMAP;
	highlight_style = STYLE_HIGHLIGHT | STYLE_THICK;

	tree_brush[0] = .25;
	tree_brush[1] = .75;
	tree_height=0;

	CMAP = new colormap("HierDisplay");

	database = false;
	SetBackend(backend);
	this->databaseInfo.activeFlag = false;
}

/**
 * Used to set the data base flag to true.
 * \param flag 
 */
void CHierDisplayInformation::SetBackend(bool flag)
{
	this->database = flag;	
}


CHierDisplayInformation::DataBaseInfo::DataBaseInfo()
{
	this->histfptr = NULL;
	this->activeFlag = false;
	#ifdef XMDV_USE_BACKEND
	this->cacheController = NULL;
	#endif
}

/**
 * This method deletes all the stuff instantiated by the backend
 * This method should be called or else there will be a very big 
 * memory leak
 * \return 
 */
CHierDisplayInformation ::DataBaseInfo::~DataBaseInfo()
{
	#ifdef XMDV_USE_BACKEND//Note not deleting cache controller
	//The following three lines are commented by Zaixian
	//It seems that these three object has been cleared in function appDisconnectDatabase 
	//of XmdvTool.cpp.  It we enable them, XmdvTool will report an error when exiting.
	//delete(this->cacheBuilder);//This also deletes the cache controller.
	//delete(this->cacheConfigStore);//This also deletes the cache config store
	//delete(this->dbConfigStore);	//This deletes the db config store
	try {
		delete(this->cacheController);
	} catch (...) {
#ifdef _DEBUG
		printf ("Failed to delete this->cacheController\n");
#endif
	}
	#endif

	if(this->histfptr != NULL)
	{
		fclose(histfptr);
		histfptr = NULL;
	}
}

/**
 * This is used to delete the cache builder 
 * configuration store and the dbConfigStore.
 * 
 * \return 
 */
void CHierDisplayInformation::DataBaseInfo::reset()
{
	#ifdef XMDV_USE_BACKEND
	delete(this->cacheBuilder);
	delete(this->dbConfigStore);	
	delete(this->cacheConfigStore);
	#endif XMDV_USE_BACKEND
	if(this->histfptr != NULL)
	{
		fclose(histfptr);
		histfptr = NULL;
	}
}