///////////////////////////////////////////////////////////// COPYRIGHT //
//
//  XmdvTool - A program for Multivariate Data Display
//
//  Copyright 1994-2003, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty. 
//  In no event will the authors be held liable for any damages arising 
//  from the use of this software.
// 
///////////////////////////////////////////////////////////// COPYRIGHT //

//////////////////////////////////////////////////////////////////////////
/// \file ChierNavDb.cpp
/// \brief This file implements the functions that evaluates the tree_contour
///	from the backend rather than the front end
/// 
///
/// \date 09\25\2003
/// \author Nishant Mehta 
//////////////////////////////////////////////////////////////////////////

/*============================== INCLUDES ==============================*/

#include <iostream>
#include <xmdv/XmdvTool.h>
#include <xmdv/CHierNavHelperDb.h>	// definition of the ... class 			
#include <xmdv/CHierHist.h>			// definition for HISTO_RES

#include <backend/db_loader.h>

/*=================================Namespaces imported====================*/
using namespace std;

/*========================== PUBLIC FUNCTIONS ==========================*/
//////////////////////////////////////////////////////////////////////////
/// \brief Constructs a new NavHelper
//////////////////////////////////////////////////////////////////////////

CHierNavHelperDb::CHierNavHelperDb()
{
	this->tree_height = 0;
	this->root_radius = -1;
}

//////////////////////////////////////////////////////////////////////////
/// \brief destroys a new NavHelper
//////////////////////////////////////////////////////////////////////////
CHierNavHelperDb::~CHierNavHelperDb()
{

}

//////////////////////////////////////////////////////////////////////////
/// \brief Finds the height of the tree
/// \return int, The height of the tree
//////////////////////////////////////////////////////////////////////////
int CHierNavHelperDb ::  FindTreeHeight()
{
	if(this->tree_height == 0)
		this->tree_height = db_get_height();

	return(this->tree_height);
}
///////////////////////////////////////////////////////////////////////
/// \brief Find the radius of the root node
/// \return double, the radius of the root node
///////////////////////////////////////////////////////////////////////
double CHierNavHelperDb :: FindRootRadius()
{
		if(this->root_radius = -1)
		{
			this->root_radius  = db_get_root_radius(g_globals.dataset->datafile);
		}
	return(this->root_radius);
}



//////////////////////////////////////////////////////////////////////////
/// \brief Populates the tree_contour array
///
/// This function evaluates fetches all the different levels from the data base
////That is the height of the tree_contour is the same as 
////as that of the length of this array. 
///
/// \param &countour_radius. an array that contains all the distinct values 
////for the radius
//////////////////////////////////////////////////////////////////////////
void CHierNavHelperDb :: FindRadiusDistribution(ARR<double> &contour_radius)
{	
	int length = db_get_height();

	//This length can also be used as the height of the tree
	this->tree_height = length;
	
	long no_of_radius_values = get_no_of_radius_values();
	
	double *temp_contour = new double[no_of_radius_values];

	long *temp_no_of_nodes = new long[no_of_radius_values];

	//This finds the all the distict values of radius and stores it in the 
	//temp_contour array
	db_radius(temp_contour,temp_no_of_nodes);
	
	//Have to find a way to initialize max-radius, to the biggest value of the radius
	double max_radius = temp_contour[0];

	//This makes the histogram from the data base and not the cluster tree
	//Note to make the histogram we do not require the cluster tree
	for(int i=0;i<HISTO_RES;i++)
		CHierHist::histo[i] = 0;


	for(int i = 0;i < no_of_radius_values;i++)
	{		
	    int pos = (int)((temp_contour[i]/max_radius)*HISTO_RES);

		if(pos >= HISTO_RES)
			pos = HISTO_RES-1;

		  CHierHist::histo[pos] = CHierHist::histo[pos] + temp_no_of_nodes[i];
	}
		
	//Evaluate the radius_contour. Note the radius_contour can be less than
	//the heignt of the tree.
	int MIN_GAP = 5;    

//    contour_radius.SetLen(0);
    contour_radius.Clear();

	long previ = 10000000;

    for (int i=CHierHist::histo.Len()-1,previ=1000000;i>=0;i--)
    {
        if (CHierHist::histo[i]==0) continue;
        if ((previ-i)<MIN_GAP) continue;

        contour_radius.InsertTail( max_radius*i/double(CHierHist::histo.Len()-1) );
        previ = i;
    }
    contour_radius.InsertTail(0.0);		
	delete[] temp_contour;
	delete[] temp_no_of_nodes;
}

//////////////////////////////////////////////////////////////////////////
/// \brief Populates the tree_contour
///
/// This function evaluates the complete tree_contour. 
/// That is for each level ot fetches all the nodes at the levels and then
/// sets the values in the tree_contour[i]. Where i corresponds to the depth 
/// of the tree you are currently working at
/// \param &tree_contour, a 2 dimensional array that stores the complete tree_contour
/// \param &tree_color, a 2 dimensional array that stores the color
/// \param contour_radius, this is used to evaluate the contour, gives all  different level
//////////////////////////////////////////////////////////////////////////

void CHierNavHelperDb :: FindTreeContour(ARR< ARR<double> > &tree_contour,
										 ARR< ARR<double> > &tree_color,
										 ARR<double> contour_radius)
{
//We use contour_radius.Len() -1 because we have appended 0 at the end.
	for(int i=0;i<contour_radius.Len();i++)
	{
		long no_of_nodes = no_of_nodes_at(contour_radius[i]);
		if(no_of_nodes != 0)
		{
				//Allocating Memory for nodes
				double *arr_radius = new double[no_of_nodes];

				double *arr_parent_radius = new double[no_of_nodes];

				int *arr_height = new int[no_of_nodes];

				get_nodes_at(contour_radius[i],arr_radius,arr_parent_radius,
							arr_height);

				//
				// If it's a leaf node, Onlyy leaf nodes have a radius of 0
				//
				if (arr_radius[i] == 0 && arr_radius[i] >= contour_radius[i]) {
					arr_radius[i] =arr_radius[i]+ 0.01; // (1.0/(double)CONTOUR_RES);
				}

				for(int j =0;j < no_of_nodes;j++)
				{
					double interp;

					if (arr_parent_radius[j] - arr_radius[j] < 0.0000001) 
						interp = 0;
					else
						interp = (contour_radius[i] - arr_radius[j]) /
								(arr_parent_radius[j] - arr_radius[j]);
					//Note we use height -1 becuase the front end interprets height of root as 0.
					double y = (arr_height[j]-1) - interp;

					double L = 	(double)j/(double) no_of_nodes;

					double R = (double)(j+1)/(double) no_of_nodes;
					
					// Find the contour segment(s) this node covers: [x0,x1]					
					double x0 = (double) 
								L*(double)CONTOUR_RES;

					double x1 = (double) 
								R*(double)CONTOUR_RES;

					int xi0 = (int) ceil(x0+.000000001);
					int xi1 = (int) floor(x1-.00000001);					

					for (int k=xi0;k<xi1;k++) {
						// Coverage that fills whole segments
						assert(tree_contour[i][k] == 0.0);
						tree_contour[i][k] = y;
						tree_color[i][k] = (L+R)*.5;
					}
					if (xi0>0 && x1>xi0) {
						// Left end of the coverage
						double coverage = (xi0-x0);
						tree_contour[i][xi0-1] += coverage*y;
						if (coverage>.9) tree_color[i][xi0-1] = (L+R)*.5;
					}
					if (xi1<CONTOUR_RES && x0<xi1) {
						// Right end of the coverage
						double coverage = (x1-xi1);
						tree_contour[i][xi1] += coverage*y;
						if (coverage>.9) tree_color[i][xi1] = (L+R)*.5;
					}
					if (xi0>=x1 && xi1<=x0) {
						// Partial coverage
						double coverage = (x1-x0);
						tree_contour[i][xi1] += coverage*y;
						if (coverage>.9) tree_color[i][xi1] = (L+R)*.5;
					}

		  }//End of inner for

		//Freeing the memory once we have calculated the contour		
		delete[] arr_radius;
		delete[] arr_parent_radius;
		delete[] arr_height;		
		}//end of if(no_of_nodes != 0)				
	}//end of outer for	
}
