
#include <math.h>
#include <xmdv/Segment.h>
#include <xmdv/Image.h>
#include <xmdv/RGB.h>
#include <xmdv/colormap.h>
#include <xmdv/XmdvTool.h>
#include <xmdv/okcdata.h>

inline int round(double x)
{	return (int)(x + 0.5);	}

Segment::Segment(int nMaxPixel, int nColumnIndex,
				 double mLeftLine, double dLeftAngle, bool bLeftLineDepX,
				 double mRightLine, double dRightAngle, bool bRightLineDepX,
				 double mDrawLine, double dDrawLineAngle, bool bDrawLineDepX,
				 bool bMirrored, int nMirrorSegY, int nMirrorSegX,
				 int nMirrorSegXY, ARR<Segment *> &segments)
				 : m_segments(segments)
{
	m_bSegCalculated = false;
	m_bMirrored = false;
	m_nMirrored = -1;
	m_nMaxPixel = nMaxPixel;
	m_nColumnIndex = nColumnIndex;
	m_mLeftLine = mLeftLine;
	m_dLeftAngle = dLeftAngle;
	m_bLeftLineDepX = bLeftLineDepX;
	m_mRightLine = mRightLine;
	m_dRightAngle = dRightAngle;
	m_bRightLineDepX = bRightLineDepX;
	m_mDrawLine = mDrawLine;
	m_dDrawLineAngle = dDrawLineAngle;
	m_bDrawLineDepX = bDrawLineDepX;
	m_bMirrored = bMirrored;
	m_nMirrorSegY = nMirrorSegY;
	m_nMirrorSegX = nMirrorSegX;
	m_nMirrorSegXY = nMirrorSegXY;
//	m_segments = segments;
//	m_norm = new double[g_globals.dataset->data_size];
//	g_globals.dataset->GetDimData(m_norm, nColumnIndex);
//	m_norm.SetLen(data.Len());
//	for (int c = 0; c < data.Len(); ++c) {
//		m_norm[c] = (data[c] - g_globals.dataset->dim_min[nColumnIndex]) /
//			(g_globals.dataset->dim_max[nColumnIndex] - g_globals.dataset->dim_min[nColumnIndex]);
//	}
	if (m_bDrawLineDepX) {
		if (m_dDrawLineAngle > 90.0 && m_dDrawLineAngle <= 270.0) {
			m_nMoveIncrementor = 1;
		} else {
			m_nMoveIncrementor = -1;
		}
	} else if (m_dDrawLineAngle > 0.0 && m_dDrawLineAngle <= 180.0) {
		m_nMoveIncrementor = 1;
	} else {
		m_nMoveIncrementor = -1;
	}
	m_colorChord = new RGBt(1.0, 1.0, 1.0);
	m_colorChordInv = new RGBt(0, 0, 0);
	resetFunctionValues();
}

Segment::~Segment()
{
	delete [] m_norm;
	delete m_colorChord;
	delete m_colorChordInv;
}

void Segment::resetFunctionValues()
{
	m_nDrawLineDirection = 1;
	changeDirection();
	m_nDrawLineT = 0;
}

void Segment::moveDrawLine()
{
	m_nDrawLineT += m_nMoveIncrementor;
}

void Segment::changeDirection()
{
	m_nDrawLineDirection = -m_nDrawLineDirection;
	if (m_bDrawLineDepX) {
		if (m_nMoveIncrementor > 0) {
			if (1 == m_nDrawLineDirection) {
				m_bFromLeftToRight = true;
			} else {
				m_bFromLeftToRight = false;
			}
		} else if (1 == m_nDrawLineDirection) {
			m_bFromLeftToRight = false;
		} else {
			m_bFromLeftToRight = true;
		}
	} else if (m_nMoveIncrementor > 0) {
		if (1 == m_nDrawLineDirection) {
			m_bFromLeftToRight = false;
		} else {
			m_bFromLeftToRight = true;
		}
	} else if (1 == m_nDrawLineDirection) {
		m_bFromLeftToRight = true;
	} else {
		m_bFromLeftToRight = false;
	}
}

int Segment::point_betw_lines(int x, int y)
{
	if (x == 0 && y == 0) {
		return 5;
	}
	int nLeftPosition = 0;
	int nRightPosition = 0;
	if (m_bLeftLineDepX) {
		if (y == (int) round(m_mLeftLine * (double) x + 0.0)) {
			return !m_bFromLeftToRight ? 6 : 7;
		}
	} else if (x == (int) round((double) y / m_mLeftLine)) {
		return !m_bFromLeftToRight ? 6 : 7;
	}
	if (m_bRightLineDepX) {
		if (y == (int) round(m_mRightLine * (double) x + 0.0)) {
			return !m_bFromLeftToRight ? 7 : 6;
		}
	} else if (x == (int) round((double) y / m_mRightLine)) {
		return !m_bFromLeftToRight ? 7 : 6;
	}
	double dAngle = 0.0;
	double dLeftAngle = m_dLeftAngle;
	double dRightAngle = m_dRightAngle;
	if (x == 0) {
		if (y < 0) {
			dAngle = 270.0;
		} else {
			dAngle = 90.0;
		}
	} else {
		double dDivide = (double) y / (double) x;
		dAngle = atan(dDivide);
		dAngle = (dAngle * 180.0) / 3.1415926535897931;
	}
	if (x < 0) {
		dAngle = 180.0 + dAngle;
	} else if (x > 0 && y < 0) {
		dAngle = 360.0 + dAngle;
	}
	if (0.0 == dAngle && (360.0 == dLeftAngle || 360.0 == dRightAngle)) {
		dAngle = 360.0;
	}
	if (m_bDrawLineDepX) {
		if (m_mDrawLine < 0) {
			if (m_nMoveIncrementor > 0) {
				if (dAngle >= 270.0)
					dAngle -= 270.0;
				else
					dAngle += 90.0;
				if (dRightAngle >= 270.0)
					dRightAngle -= 270.0;
				else
					dRightAngle += 90.0;
				if (dLeftAngle >= 270.0)
					dLeftAngle -= 270.0;
				else
					dLeftAngle += 90.0;
			} else if (0.0 == dAngle) {
				dAngle = 360.0;
			}
		} else if (m_nMoveIncrementor < 0) {
			if (dAngle >= 180.0)
				dAngle -= 180.0;
			else
				dAngle += 180.0;
			if (dRightAngle >= 180.0)
				dRightAngle -= 180.0;
			else
				dRightAngle += 180.0;
			if (dLeftAngle >= 180.0)
				dLeftAngle -= 180.0;
			else
				dLeftAngle += 180.0;
		}
	} else if (m_mDrawLine < 0) {
		if (m_nMoveIncrementor > 0) {
			if (dAngle >= 180.0)
				dAngle -= 180.0;
			else
				dAngle += 180.0;
			if (dRightAngle >= 180.0)
				dRightAngle -= 180.0;
			else
				dRightAngle += 180.0;
			if (dLeftAngle >= 180.0)
				dLeftAngle -= 180.0;
			else
				dLeftAngle += 180.0;
		}
	} else if (90.0 != dRightAngle) {
		if (dAngle >= 90.0)
			dAngle -= 90.0;
		else
			dAngle += 270.0;
		if (dRightAngle >= 90.0)
			dRightAngle -= 90.0;
		else
			dRightAngle += 270.0;
		if (dLeftAngle >= 90.0)
			dLeftAngle -= 90.0;
		else
			dLeftAngle += 270.0;
	}
	if ((dAngle - dLeftAngle) * (dAngle - dRightAngle) > 0) {
		if (m_bDrawLineDepX) {
			if (m_nMoveIncrementor > 0) {
				if (1 == m_nDrawLineDirection)
					return dAngle >= dRightAngle ? 3 : 4;
				return dAngle <= dLeftAngle ? 3 : 4;
			}
			if (1 == m_nDrawLineDirection)
				return dAngle <= dLeftAngle ? 3 : 4;
			return dAngle >= dRightAngle ? 3 : 4;
		}
		if (m_nMoveIncrementor > 0) {
			if (1 == m_nDrawLineDirection)
				return dAngle <= dLeftAngle ? 3 : 4;
			return dAngle >= dRightAngle ? 3 : 4;
		}
		if (1 == m_nDrawLineDirection)
			return dAngle >= dRightAngle ? 3 : 4;
		return dAngle <= dLeftAngle ? 3 : 4;
	} else {
		return 2;
	}
}

void Segment::compute_next_point(int &x, int &y)
{
	if (m_bDrawLineDepX) {
		x += m_nDrawLineDirection;
		y = (int) round(m_mDrawLine * (double) x + (double) m_nDrawLineT);
	} else {
		y += m_nDrawLineDirection;
		x = (int) round(m_mDrawLine * (double) y + (double) m_nDrawLineT);
	}
}

/*virtual*/ void Segment::Paint(Image *image,double samplingFactor /*=1.0*/,bool hilite /*=false*/)
{
//	m_cmap = cmap;
	if (m_bMirrored) {
		return;
	}
	int nLastValidT = 0, x = 0, y = 0;
	try {
		int nCounter = initial_pixels(image, x, y);
		m_nLastDrawLineLength = 0;
		nCounter = left_pixels(image, nCounter, x, y);
		nLastValidT = m_nDrawLineT;
		for (int nPosition = point_betw_lines(x, y);
			 4 != nPosition;
			 nPosition = point_betw_lines(x, y)) {
			compute_next_point(x, y);
		}

		if (NOT_MIRRORED != m_nMirrorSegY) {
			m_segments[m_nMirrorSegY]->m_nMirrored = Y_MIRRORED;
		}
		if (NOT_MIRRORED != m_nMirrorSegX) {
			m_segments[m_nMirrorSegX]->m_nMirrored = X_MIRRORED;
		}
		if (NOT_MIRRORED != m_nMirrorSegXY) {
			m_segments[m_nMirrorSegXY]->m_nMirrored = XY_MIRRORED;
		}
		moveDrawLine();
		resetFunctionValues();
	} catch (...) {
		printf("Something wrong!");
	}
}

int Segment::initial_pixels(Image *image, int &x, int &y)
{
	int nNumOfPixelsOnRow = 0;
	int nNumPainted = 0;
	int nPosition = 0;
	int nZoomFactor = 0;
	do {
		if (nNumOfPixelsOnRow > 4)
			break;
		nNumOfPixelsOnRow = 0;
		for (nPosition = point_betw_lines(x, y);
			 2 == nPosition && nNumPainted < m_nMaxPixel;
			 nPosition = point_betw_lines(x, y)) {
			nNumOfPixelsOnRow++;
			nNumPainted++;
			writeToImage(image, x, -y, nNumPainted-1);
			if (-1 != m_nMirrorSegY) {
				writeToImage(image, -x, -y, nNumPainted-1);
			}
			if (-1 != m_nMirrorSegX) {
				writeToImage(image, x, y, nNumPainted-1);
			}
			if (-1 != m_nMirrorSegXY) {
				writeToImage(image, -x, y, nNumPainted-1);
			}
			compute_next_point(x, y);
		}

		if (nNumPainted >= m_nMaxPixel)
			break;
		for (;
			 4 != nPosition && 3 != nPosition;
			 nPosition = point_betw_lines(x, y)) {
			compute_next_point(x, y);
		}

		while (2 != nPosition) {
			moveDrawLine();
			compute_next_point(x, y);
			changeDirection();
			for (nPosition = point_betw_lines(x, y);
				 3 == nPosition;
				 nPosition = point_betw_lines(x, y))
				compute_next_point(x, y);

			while (2 != nPosition && 4 != nPosition) {
				compute_next_point(x, y);
				nPosition = point_betw_lines(x, y);
			}
		}
	} while (true);
	return nNumPainted;
}

void Segment::writeToImage(Image *image, int x, int y, int index)
{
/*	int cx = image->Width()/2, cy = image->Height()/2;
	okcdata *dataset = g_globals.dataset;
	QueryDist &query_dist = dataset->flat_query;

//	double brushedOpacity = 1.0, unbrushedOpacity = 1.0, opacity = 1.0;
//	if (hier_info) {
//		samplingFactor = 1.0f;
//		if (hier_info->fade_factor>0) 
//			unbrushedOpacity = 1-hier_info->fade_factor;
//		else if (hier_info->fade_factor<0) 
//			brushedOpacity = 1+hier_info->fade_factor;
//	}
//	RGBt iface_color(interface_color[BG]);

//	double start = g_globals.GetRandomNum() * (1-samplingFactor);
//	double end = start + samplingFactor;

	colormap *cmap;
	int idx = index*(dataset->all_dims+1)+m_nColumnIndex;
	double val = query_dist.distances[idx];
	int aaa = 0;

//	if (query_dist.randomNums[index] < start ||
//		query_dist.randomNums[index] > end)
//	{
//		return;
//	}

	double min, max;

	if (m_nColumnIndex == 0) {
		if (val < query_dist.brush_level) {
			min = query_dist.dist_min;
			max = query_dist.brush_level;
			cmap = g_globals.pixelBrushedCMap;
//			opacity = brushedOpacity;
			aaa = 1;
		} else {
			min = query_dist.brush_level;
			max = query_dist.dist_max;
			cmap = g_globals.pixelUnbrushedCMap;
//			opacity = unbrushedOpacity;
			aaa = 2;
		}
	} // if (dim == 0) 
	else {
		Brush &brush = *g_globals.curr_brush;
		// negative values must be below brush center
		if (val <= 0) 
		{
			if (val <= brush.Min(m_nColumnIndex-1) - brush.pos[m_nColumnIndex-1]) {
				max = brush.Min(m_nColumnIndex-1) - brush.pos[m_nColumnIndex-1];
				min = dataset->all_dim_min[m_nColumnIndex-1] - brush.pos[m_nColumnIndex-1];
				cmap = g_globals.pixelUnbrushedCMap;
//				opacity = unbrushedOpacity;
				aaa = 3;
			} else {
				max = 0;
				min = brush.Min(m_nColumnIndex-1) - brush.pos[m_nColumnIndex-1];
				cmap = g_globals.pixelBrushedCMap;
//				opacity = brushedOpacity;
				aaa = 4;
			}
		} // if (val <= 0) 
		else {
			if (val <= brush.Max(m_nColumnIndex-1) - brush.pos[m_nColumnIndex-1]) {
				min = 0;
				max = brush.Max(m_nColumnIndex-1) - brush.pos[m_nColumnIndex-1];
				cmap = g_globals.pixelBrushedCMap;
//				opacity = brushedOpacity;
				aaa = 5;
			} 
			else {
				min = brush.Max(m_nColumnIndex-1) - brush.pos[m_nColumnIndex-1];
				max = dataset->all_dim_max[m_nColumnIndex-1] - brush.pos[m_nColumnIndex-1];
				cmap = g_globals.pixelUnbrushedCMap;
//				opacity = unbrushedOpacity;
				aaa = 6;
			}
		}
	}

	assert(val >= min && val <= max);

	double norm_val = (val - min) / (max - min);
	unsigned long color = (*cmap)[int(norm_val * (cmap->Len() - 1))];
	int iii = int((norm_val * (cmap->Len() - 1)) + 0.5);
	RGBt rgbColor = (*cmap)[iii];
//	rgbColor.BlendTo(iface_color, opacity);
	image->WritePixel(cx+x, cx+x, rgbColor.GetColor());
//	image->WritePixel(cx+x, cx+y, (unsigned char) (color >> 16) & 0xff,
//								(unsigned char) (color >> 8) & 0xff,
//								(unsigned char) color & 0xff);
*/}

int Segment::left_pixels(Image *image, int nCounter, int &x, int &y)
{
	int nPosition = 0;
	int nZoomFactor = 0;
	int nNewLineLength = 0;
	int nRunTo = 0;
	do {
		if (nCounter >= m_nMaxPixel)
			break;
		nPosition = point_betw_lines(x, y);
		nNewLineLength = 0;
		nRunTo = m_nLastDrawLineLength - 2;
		for (;
			 2 == nPosition && nCounter < m_nMaxPixel;
			 nPosition = point_betw_lines(x, y))
			do {
				nCounter++;
				writeToImage(image, x, -y, nCounter-1);
				if (-1 != m_nMirrorSegY) {
					writeToImage(image, -x, -y, nCounter-1);
				}
				if (-1 != m_nMirrorSegX) {
					writeToImage(image, x, y, nCounter-1);
				}
				if (-1 != m_nMirrorSegXY) {
					writeToImage(image, -x, y, nCounter-1);
				}
				nNewLineLength++;
				compute_next_point(x, y);
			} while (nNewLineLength <= nRunTo && nCounter < m_nMaxPixel);

		m_nLastDrawLineLength = nNewLineLength;
		if (nCounter >= m_nMaxPixel)
			break;
		for (; 4 != nPosition; nPosition = point_betw_lines(x, y)) {
			compute_next_point(x, y);
		}

		moveDrawLine();
		compute_next_point(x, y);
		changeDirection();
		for (nPosition = point_betw_lines(x, y);
			 3 == nPosition;
			 nPosition = point_betw_lines(x, y))
			compute_next_point(x, y);

		while (2 != nPosition) {
			compute_next_point(x, y);
			nPosition = point_betw_lines(x, y);
		}
	} while (true);
	return nCounter;
}

bool Segment::Contains(float x, float y)
{
	// TODO: Add code here
	return false;
}