//modified by Jing Yang after 7/6/2001

// SYSTEM INCLUDES
//
#include <cstdio>
#include <cstring>
#include <cstdlib>
#include <cmath>
#include <search.h>
#include <limits>
#include <ctime>



// PROJECT INCLUDES
//
#include <xmdv/okcdata.h>
#include <xmdv/macros.h>
#include <xmdv/brush.h>
#include <xmdv/brush_oper.h>
#include <xmdv/CScreenDistortion.h>
#include <xmdv/XmdvTool.h>
#include <xmdv/ARR.h>
#include <xmdv/CDimReducer.h>
//added by debbie
#include <xmdv/brush_oper.h>
#include <xmdv/OrderDimension.h>


#ifdef USE_META
#include <string>
#include <fstream>
#include <meta/Meta.h>
#endif


double relfacBrushMin, relfacBrushMax;

okcdata::okcdata()
	: data_size(0), all_dims(0), dims(0), datafile(NULL)
{
#ifdef USE_META
	meta = 0;
	hasMeta = false;
#endif // USE_META

	basex = NULL;
	basey = NULL;
	brushed_time = NULL;
	brushMgr = NULL;
}

okcdata::~okcdata()
{
	for (int i=0; i<all_names.Len(); ++i)
	{
		delete [] all_names[i];
	}
	if (basex) delete [] basex;
	if (basey) delete [] basey;
	if (brushed_time) delete [] brushed_time;
	if (brushMgr) delete brushMgr;
}

static const char * read_till(FILE *fp,
							  const char *delimiters = "\n",
							  char *which_delimiter = 0)
{
	static char str[500];
	int k = 0;
	while (1) {
		char c = fgetc(fp);
		char *ptr = strchr(delimiters, c);
		if (ptr) {
			str[k] = '\0';
			if (which_delimiter)
				*which_delimiter = *ptr;
			break;
		}
		str[k++] = c;
	}

	/*if ( strchr(str, ' ') )
	{
		char str1[500];
		sprintf(str1, "{%s}", str);
		strcpy(str, str1);
	}*/
	return str;
}



////////////////////////////////////////



// Read in data file: Format is:
// number_of_dimensions number_of_samples (int int)
// dimension_name for each dimension (string)
// minimum maximum cardinality for each dimension (double double int)
// data_samples (doubles)
//


bool okcdata::Read(const char *fname, okcdata::progress p/* = NULL*/)
{
	FILE *fp = fopen(fname, "r");
	if (fp == NULL) {
		fprintf(stderr, "Data file %s doesn't exist\n", fname);
		return false;
	}

	datafile = strdup(fname);

#ifdef USE_META
	// Get the meta filename from the okc file name
	//
	std::string okcFilename(fname);
	char *temp = strrchr(fname, '.');
	std::string metaFilename = okcFilename.substr(0, temp - fname + 1) +
							   "meta";
	std::ifstream is(metaFilename.c_str());
	if (is) {
		meta = MetaFactory::instance()->createMeta();
		if (meta == NULL)
			std::cerr << "Meta NOT instantiated successfully" << std::endl;
		hasMeta = meta->unmarshal(is, 0);
	} else {
		hasMeta = false;
	}
#endif // USE_META

	// get length of file for progress bar
	fseek(fp, 0, SEEK_END);
	long length	= ftell(fp);
	fseek(fp, 0, SEEK_SET);

	// initialize progress values
	long progress	= 0;
	int x = 0, prevx = -1;

	// Get number of dimensions, and number of records
	//
	if (fscanf(fp, "%d %d\n", &all_dims, &data_size) <= 0) {
		printf("Error reading first line of file %s!\n", fname);
		printf("Try loading another file!\n");
		return false;
	}

	if (data_size < 0) {
		printf("Number of data items is negative (%d)!\n", data_size);
		printf("Try loading another file!\n");
		return false;
	}

	// Read in names of dimensions, and if necessary, nominal
	// encodings of values.
	//
	all_names.SetLen(all_dims);
	all_nominal_fields.SetLen(all_dims);

	int i;
	for (i = 0; i < all_dims; i++) 
	{
		// get the progress of our parsing
		progress = ftell(fp);
		x = (int) (((float) progress / length) * 75);

		// output it if it's significantly different from last 
		// displayed progress marker
		if (x != prevx)
		{
			if (p != NULL)
				(*p)(x);
		}
		prevx = x;

		char endchar;
		const char *dname = read_till(fp, "(\n", &endchar);
		all_names[i] = new char[strlen(dname) + 4];
		strcpy(all_names[i], dname);

		NominalField *ndata = 0;
		if (endchar == '(') {
/*			ndata = new NominalField;
			ndata->range_min = atoi(read_till(fp,",",&endchar));
			ndata->range_max = atoi(read_till(fp,"),\n",&endchar));
			ndata->values.SetLen(ndata->range_max-ndata->range_min+1);
			int n;
			for (n=0;n<ndata->values.Len();n++) {
				ndata->values[n] = 0;
			}
			if (endchar != ')') 
			{
				for (n=ndata->range_min; n<=ndata->range_max; n++) {
					const char *str = read_till(fp,",)\n",&endchar);
					ndata->values[n-ndata->range_min] = strdup(str);
				}
			}
*/
			read_till(fp, "\n");
		}

		all_nominal_fields[i] = ndata;
	}

	all_dim_min.SetLen(all_dims);
	all_dim_max.SetLen(all_dims);
	cg_dim_min.SetLen(all_dims);
	cg_dim_max.SetLen(all_dims);
	all_cardinality.SetLen(all_dims);
	all_weight.SetLen(all_dims);
	on.SetLen(all_dims);

	for (i = 0; i < all_dims; i++) 
	{
		// get the progress of our parsing
		progress = ftell(fp);
		x = (int) (((float) progress / length) * 75);

		// output it if it's significantly different from last 
		// displayed progress marker
		if (x != prevx)
		{
			if (p != NULL)
				(*p)(x);
		}
		prevx = x;

		if (fscanf(fp,
				   "%lf %lf %d",
				   &all_dim_min[i],
				   &all_dim_max[i],
				   &all_cardinality[i]) != 3) {
			printf("failed in reading dimension ranges.\n");
			return false;
		}

	    all_weight[i] = 1;

		cg_dim_min[i] = all_dim_min[i];
		cg_dim_max[i] = all_dim_max[i];

		// Need to handle case when min = max (separate them a bit)
		//
		if (all_dim_min[i] == all_dim_max[i]) {
			if (all_dim_min[i] == 0.0) {
				all_dim_min[i] = -1.0;
				all_dim_max[i] = 1.0;
			} else {
				if (all_dim_min[i] < 0.0) {
					all_dim_min[i] = 2. * all_dim_min[i];
					all_dim_max[i] = 0.0;
				} else {
					all_dim_max[i] = 2. * all_dim_min[i];
					all_dim_min[i] = 0.0;
				}
			}
		} else {
			// Add a bit of space to the bounds to avoid boundary problems
			//
			const double SPACE = .05; // extra space for bounds to avoid landing on edges

			double dim_diff = (all_dim_max[i] - all_dim_min[i]) * SPACE;
			all_dim_max[i] += dim_diff;
			all_dim_min[i] -= dim_diff;
		}

		on[i] = true;
	}

	data_buf.SetLen(all_dims * data_size);

	// PAUL: io_org
	// Use less polling
	int k = 0;
	for (i = 0; i < data_size; i++) {

		// PAUL: io_org
		// get the progress of our parsing
        if (i % 1000 == 0)
		{
			progress = ftell(fp);
			x = (int) (((float) progress / length) * 75);

			// output it if it's significantly different from last 
			// displayed progress marker
			if (x != prevx)
			{
				if (p != NULL)
					(*p)(x);
			}
			prevx = x;
		}

		for (int j = 0; j < all_dims; j++,k++) {


			if (fscanf(fp, "%lf", &data_buf[k]) != 1) {
				printf("failed in reading data, line %d, item %d.\n", i, j);
				return false;
			}
		}
	}

	fclose(fp);

// PAUL: timing
	/*
DWORD dwEnd = GetTickCount();
char buffer[52];
int len = sprintf(buffer, "Ticks = %d\n",
					dwEnd - dwStart);
printf(buffer);

dwStart = GetTickCount();
*/

	mark.SetLen(data_size);
	color.SetLen(data_size);

	names.SetLen(all_dims);
	dim_min.SetLen(all_dims);
	dim_max.SetLen(all_dims);
	nominal_fields.SetLen(all_dims);
	axis_color.SetLen(all_dims);
	org_dim_min.SetLen(all_dims);
	org_dim_max.SetLen(all_dims);
	cardinality.SetLen(all_dims);
	tempCardinality.SetLen(all_dims);
	weight.SetLen(all_dims);
	tempWeight.SetLen(all_dims);
	//used for reordering dimensions
	map_file_to_buf.SetLen(all_dims);
	map_buf_to_file.SetLen(all_dims);
	for (i = 0; i < all_dims; i++) {
		map_file_to_buf[i] = i;
		map_buf_to_file[i] = i;
	}

	for (i = 0; i < all_dims; i++) {
		axis_color[i] = interface_color[GRID1];
	}

	//used for glyph placement
	basex = new double[data_size];
	basey = new double[data_size];

	//start_jing_animate_brush
	brushed_time = new double[data_size];
	animate_brush = false;  			  
	for (i = 0; i < data_size; i++)
		brushed_time[i] = -1;
	//end_jing_animate_brush

	SetFieldOnOff(-1, true);

    Placement_Original_Glyph();

	// Anil [5/14/2004] - allocate space for query distances
//	query_dist.SetLen((all_dims+1)*data_size);
	//

	//start_debbie_reorder

	//end_debbie_reorder


	// PAUL: draw_partial
	// Give every data item a random variable
	rand_values.SetLen(data_size);
	for (int i = 0; i < data_size; ++i)
		rand_values[i] = rand();

// PAUL: timing
	/*
dwEnd = GetTickCount();
len = sprintf(buffer, "Ticks (Rest) = %d\n",
					dwEnd - dwStart);
printf(buffer);
*/

	// PAUL: brush_org
	// Initialize the BrushMgr
    brushMgr = new BrushMgr(*this);	

	return true;
}


//read from file for reordering.
void okcdata::Read2()
{
	FILE *fp = fopen(datafile, "r");
	if (fp == NULL) {
		fprintf(stderr, "Data file %s doesn't exist\n", datafile);
		return;
	}

	// Get number of dimensions, and number of records
	//
	int all_dims2, data_size2;
	if (fscanf(fp, "%d %d\n", &all_dims2, &data_size2) <= 0) {
		printf("Error reading first line of file %s!\n", datafile);
		printf("Try loading another file!\n");
		all_dims = 0;
		data_size = 0;
	}

	if ((all_dims2 != all_dims) || (data_size2 != data_size)) {
		return;
	}

	// Read in names of dimensions, and if necessary, nominal
	// encodings of values.
	//
	int i, loopi;
	for (loopi = 0; loopi < all_dims; loopi++) {
		i = map_file_to_buf[loopi];
		char endchar;
		const char *dname = read_till(fp, "(\n", &endchar);
		all_names[i] = new char[strlen(dname) + 4];
		strcpy(all_names[i], dname);

		NominalField *ndata = 0;
		if (endchar == '(') {
			ndata = new NominalField;
			ndata->range_min = atoi(read_till(fp, "),\n", &endchar));
			ndata->range_max = atoi(read_till(fp, "),\n", &endchar));
			ndata->values.SetLen(ndata->range_max - ndata->range_min + 1);

			int n;
			for (n = 0; n < ndata->values.Len(); n++) {
				ndata->values[n] = 0;
			}

			if (endchar != ')') {
				for (n = ndata->range_min; n <= ndata->range_max; n++) {
					const char *str = read_till(fp, ",)\n", &endchar);
					ndata->values[n - ndata->range_min] = strdup(str);
				}
			}

			read_till(fp, "\n");
		}

		all_nominal_fields[i] = ndata;
	}

	for (loopi = 0; loopi < all_dims; loopi++) {
		i = map_file_to_buf[loopi];
		if (fscanf(fp,
				   "%lf %lf %d",
				   &all_dim_min[i],
				   &all_dim_max[i],
				   &all_cardinality[i]) != 3) {
			printf("failed in reading dimension ranges.\n");
			exit(-1);
		}
		all_weight[i] = 1;
		cg_dim_min[i] = all_dim_min[i];
		cg_dim_max[i] = all_dim_max[i];

		// Need to handle case when min = max (separate them a bit)
		//
		if (all_dim_min[i] == all_dim_max[i]) {
			if (all_dim_min[i] == 0.0) {
				all_dim_min[i] = -1.0;
				all_dim_max[i] = 1.0;
			} else {
				if (all_dim_min[i] < 0.0) {
					all_dim_min[i] = 2. * all_dim_min[i];
					all_dim_max[i] = 0.0;
				} else {
					all_dim_max[i] = 2. * all_dim_min[i];
					all_dim_min[i] = 0.0;
				}
			}
		} else {
			// Add a bit of space to the bounds to avoid boundary problems
			//
			const double SPACE = .05; // extra space for bounds to avoid landing on edges

			double dim_diff = (all_dim_max[i] - all_dim_min[i]) * SPACE;
			all_dim_max[i] += dim_diff;
			all_dim_min[i] -= dim_diff;
		}
	}

	int base, k;
	for (i = 0; i < data_size; i++) {
		base = i * all_dims;
		for (int j = 0; j < all_dims; j++) {
			k = base + map_file_to_buf[j];
			if (fscanf(fp, "%lf", &data_buf[k]) != 1) {
				printf("failed in reading data, line %d, item %d.\n", i, j);
			}
		}
	}
	fclose(fp);
}


void okcdata::WriteBrushed(const char *fname)
{
	if (strcmp(fname, "") == 0) {
		return;
	}
	int i, j;

	int count = 0;

	ARR<double> data;
	ARR<double> max;
	ARR<double> min;

	data.SetLen(dims);
	max.SetLen(dims);
	min.SetLen(dims);

	bool brushed;

	for (i = 0; i < dims; i++) {
		max[i] = g_globals.curr_brush->Min(i);
		min[i] = g_globals.curr_brush->Max(i);
	}

	for (i = 0; i < data_size; i++) {
		GetData(data, i);
		brushed = true;
		for (j = 0; j < dims; j++) {
			if (data[j] < g_globals.curr_brush->Min(j))
				brushed = false;
			if (data[j] > g_globals.curr_brush->Max(j))
				brushed = false;
		}
		if (brushed) {
			for (j = 0; j < dims; j++) {
				if (data[j] > max[j])
					max[j] = data[j];
				if (data[j] < min[j])
					min[j] = data[j];
			}
			count ++;
		}
	}



	FILE *fp = fopen(fname, "w");

	// Get number of dimensions, and number of records
	//
	fprintf(fp, "%d %d\n", dims, count);

	// Read in names of dimensions, and if necessary, nominal
	// encodings of values.
	//
	for (i = 0; i < dims; i++) {
		fprintf(fp, "%s\n", names[i]);
	}

	for (i = 0; i < dims; i++) {
		fprintf(fp, "%lf %lf %d\n", min[i], max[i], cardinality[i]);
	}

	for (i = 0; i < data_size; i++) {
		GetData(data, i);
		brushed = true;
		for (j = 0; j < dims; j++) {
			if (data[j] < g_globals.curr_brush->Min(j))
				brushed = false;
			if (data[j] > g_globals.curr_brush->Max(j))
				brushed = false;
		}
		if (brushed) {
			for (j = 0; j < dims; j++) {
				fprintf(fp, "%lf ", data[j]);
			}
			fprintf(fp, "\n");
		}
	}

	fclose(fp);
}

void okcdata::GetAllData(ARR<double> &buf, int line)
{
	int j = 0;
	int base = all_dims *line;
	for (int i = 0; i < all_dims; i++)
		buf[j++] = data_buf[base + i];
} 

void okcdata::GetData(ARR<double> &buf, int line)
{
	// PAUL: lang_opt
	// Make all_data static so that it's not new[]'d every time it is called
	static ARR<double> all_data;	// static, so all_data has both a global lifetime and local scope
	all_data.SetLen(all_dims);

	int j = 0;
	int base = all_dims *line;
	for (int i = 0; i < all_dims; i++)
		all_data[j++] = data_buf[base + i];
	MapData(all_data, buf);
} 

void okcdata::GetDimData(double *& buf, int dim)
{
	assert(buf != NULL);
	int idx = dim;
	// PAUL: reorder_math
	double inv_dim_range = 1 / (dim_max[dim] - dim_min[dim]);
	for (int i = 0; i < data_size; ++i, idx += all_dims) {
//		idx = i * all_dims + dim;
//		buf[i] = data_buf[idx];
		buf[i] = (data_buf[idx] - dim_min[dim]) * inv_dim_range;
	}
}

//void okcdata::GetDimDist(double *&buf,int dim)
//{
//	assert(buf != NULL);
//	int idx = dim;
//	Brush &brush = g_globals.brushes[0];
//	double min, max;
//	for (int i = 0; i < data_size; ++i, idx += all_dims+1) 
//	{
//		if (dim == 0) {
//			if (query_dist[idx] < brush_level) {
//				min = rel_factor_min;
//				max = brush_level;
//			} else {
//				min = brush_level;
//				max = rel_factor_max;
//			}
//		} else {
//			if (query_dist[idx] <= 0) 
//			{
//				if (query_dist[idx] <= brush.Min(dim-1) - brush.pos[dim-1]) {
//					min = brush.Min(dim-1) - brush.pos[dim-1];
//					max = all_dim_min[dim-1] - brush.pos[dim-1];
//				} else {
//					min = 0;
//					max = brush.Min(dim-1) - brush.pos[dim-1];
//				}
//			} else {
//				if (query_dist[idx] <= brush.Max(dim-1) - brush.pos[dim-1]) {
//					min = 0;
//					max = brush.Max(dim-1) - brush.pos[dim-1];
//				} else {
//					min = brush.Max(dim-1) - brush.pos[dim-1];
//					max = all_dim_max[dim-1] - brush.pos[dim-1];
//				}
//			}
//		}
//		buf[i] = (query_dist[idx] - min) / (max - min);
//	}
//}

void okcdata::GetDataWithDimClusterExtends(ARR<double> &mean,
										   ARR<double> &max,
										   ARR<double> &min,
										   int line)
{
	// PAUL: lang_opt
	// make all_data static
	static ARR<double> all_data;
	all_data.SetLen(all_dims);

	int j = 0;
	int base = all_dims *line;
	for (int i = 0; i < all_dims; i++)
		all_data[j++] = data_buf[base + i];
	MapDataWithDimClusterExtends(all_data, mean, max, min);
} 


#include <xmdv/brush.h>

void okcdata::SaveBrushes()
{
	for (int br = 0; br < MAXBRUSH; br++) {
		Brush *all_brush = &g_globals.all_brushes[br];
		Brush *brush = &g_globals.brushes[br];
		/*  	  //test
					for ( int i = 0; i < all_dims; i++ )
					{
						all_brush->pos[i] = (all_dim_min[i] + all_dim_max[i]) / 2.;
						all_brush->size[i] = all_dim_max[i] - all_dim_min[i];
						all_brush->bound_offset[i] = (all_dim_max[i] - all_dim_min[i]) / 10.0;
					}  */


		// PAUL: brush_org
		// Ripped from okcdata::ReverseMapData
		ReverseMapBrush(brush, all_brush);
		
/*
		ReverseMapData(brush->pos, all_brush->pos);
		ReverseMapData(brush->size, all_brush->size);
		ReverseMapData(brush->bound_offset, all_brush->bound_offset);
*/
	}
}

// PAUL: brush_org
void okcdata::ReverseMapBrush(Brush * brush, Brush * all_brush)
{
	// Ripped from okcdata::ReverseMapData
	int	i, file_i;
	int	count = 0;
	for (i = 0; i < all_dims; i++)
	{
		if (on[i])
		{
			file_i = map_buf_to_file[i];
			all_brush->SetTriple(file_i, 
				brush->Pos(count), 
				brush->Size(count), 
				brush->BoundOffset(count));
			count++;
		}
	}
}


void okcdata::MapBrush(const Brush * all_brush, Brush * brush)
{
	// PAUL: lang_org brush_org
	// Static makes it global in lifetime but local in scope
	// Use indices rather than actual data
	static ARR<int>	temp;
	temp.SetLen(all_dims);
	int		i, loopi;
	for (loopi = 0; loopi < all_dims; loopi++)
	{
		i = map_file_to_buf[loopi];
		temp[i] = loopi; // PAUL: move the index rather than the data
	}

	static ARR<int>	temp1;
	temp1.SetLen(all_dims);
	int		count = 0;
	for (i = 0; i < all_dims; i++)
	{
		if (on[i])
		{
			temp1[count] = temp[i];
			count++;
		}
	}

	if ((g_globals.dim_reducer != NULL) && (g_globals.dim_reducer->status == APPLY))
	{
		// PAUL: ?
		g_globals.dim_reducer->MapData(temp1, all_brush, brush);
	}
	else
	{
		int new_index;
		for (i = 0; i < count; i++)
		{
            new_index = temp1[i];
			brush->SetTriple(i, all_brush->Pos(new_index),
								all_brush->Size(new_index),
								all_brush->BoundOffset(new_index));
		}
	}
}

void okcdata::CountDims()
{
	dims = 0;
	for (int i = 0; i < all_dims; i++)
		if (on[i]) {
			dims++;
		}
}

void okcdata::SetFieldOnOff(int which, bool state)
{
	if (which >= 0) {
		SaveBrushes();
		on[which] = state;
	}

	dims = 0;
	for (int i = 0; i < all_dims; i++)
		if (on[i]) {
			dims++;
		}

	one_over_dims = 1.0 / dims;

	MapData(all_dim_max, dim_max);
	MapData(all_dim_min, dim_min);
	MapData(all_dim_max, org_dim_max);
	MapData(all_dim_min, org_dim_min);
	MapData(all_cardinality, cardinality);
	MapData(all_cardinality, tempCardinality);
	MapData(all_weight, weight);
	MapData(all_weight, tempWeight);
	MapData(all_names, names);

	//Commented by Zaixian
	/*char * p = all_names[0];

	names[0] = all_names[0];
	names[1] = all_names[1];
	names[2] = all_names[2];
	names[3] = all_names[3];
	*/
	MapData(all_nominal_fields, nominal_fields);

	if (which >= 0) {
		for (int br = 0; br < MAXBRUSH; br++) {
			Brush *all_brush = &g_globals.all_brushes[br];
			Brush *brush = &g_globals.brushes[br];

			// PAUL: brush_org
			MapBrush(all_brush, brush);
			//MapData(all_brush->pos, brush->pos);
			//MapData(all_brush->size, brush->size);
			//MapData(all_brush->bound_offset, brush->bound_offset);
		}
	}
	g_globals.screen_distortion->Reset();

	if (which >= 0 && g_globals.dim_reducer != NULL) {
		delete g_globals.dim_reducer;
		g_globals.dim_reducer = NULL;
	}
}

void okcdata::RemapAllData()
{
	MapData(all_dim_max, dim_max);
	MapData(all_dim_min, dim_min);
	MapData(all_dim_max, org_dim_max);
	MapData(all_dim_min, org_dim_min);
	MapData(all_cardinality, cardinality);
	MapData(all_cardinality, tempCardinality);
	MapData(all_weight, weight);
	MapData(all_weight, tempWeight);
	MapData(all_names, names);
	MapData(all_nominal_fields, nominal_fields);

	for (int br = 0; br < MAXBRUSH; br++) {
		Brush *all_brush = &g_globals.all_brushes[br];
		Brush *brush = &g_globals.brushes[br];
		// PAUL: brush_org
		MapBrush(all_brush, brush);
		//MapData(all_brush->pos, brush->pos);
		//MapData(all_brush->size, brush->size);
		//MapData(all_brush->bound_offset, brush->bound_offset);
	}
	g_globals.screen_distortion->Reset();
}


//reorder the dimensions. Which indicate which dimension need to be forward or backward. If forward == 1, 
// then move this dimension forward. Else move this dimension backward.
void okcdata::ReorderDimension(int which, int forward)
{
	SaveBrushes();

	int temp_bool;
	int temp_int;

	//neighbor indicates that dimension need to be exchange with "which" dimension.
	int neighbor;

	if (forward == 1) {
		if (which == 0) {
			neighbor = all_dims - 1;
		} else {
			neighbor = which - 1;
		}
	} else {
		if (which == all_dims - 1) {
			neighbor = 0;
		} else {
			neighbor = which + 1;
		}
	}

	temp_int = map_buf_to_file[neighbor];
	map_buf_to_file[neighbor] = map_buf_to_file[which];
	map_buf_to_file[which] = temp_int;

	map_file_to_buf[map_buf_to_file[which]] = which;
	map_file_to_buf[map_buf_to_file[neighbor]] = neighbor;

	temp_bool = on[neighbor];
	on[neighbor] = on[which];
	on[which] = temp_bool;

	RemapAllData();

	if (which >= 0 && g_globals.dim_reducer != NULL) {
		delete g_globals.dim_reducer;
		g_globals.dim_reducer = NULL;
	}
}

// Added by Debbie
void okcdata::Similarity_Based_Reorder()  
{
    SaveBrushes();

	int i,j;
	int set_size;
	double **datamatrix;
	ARR<double> data(dims);
	
	datamatrix = new double *[data_size];
	for(i = 0; i < data_size; i++)
		datamatrix[i] = new double [dims];

	// parameters is defined in OrderDimension.h

	if (g_globals.reorderParameters->brushed == 0 )
	{
		for(i = 0; i < data_size; i++)
		{
			GetData(data,i);
			for(j = 0; j< dims; j++)
			{
				datamatrix[i][j] = data[map_file_to_buf[j]];
			}
		}
		set_size = data_size;
	}
	else 
	{
		set_size = 0;
		for(i = 0; i < data_size; i++)
		{
			GetData(data,i);
			if(BrushOperCheckPoint(data,0))
			{
				for(j = 0; j < dims; j++)
					datamatrix[set_size][j] = data[map_file_to_buf[j]];
				set_size++;
			}
		}
		
	}

	if( set_size <= 1)
		return;

	ARR<int> orderarray(all_dims);			// initialize the order array with the order in map_buf_to_file
	for(i = 0; i < all_dims; i++)
		orderarray[i] = i;

	
	similarity_reorder(set_size, dims, datamatrix, orderarray, *g_globals.reorderParameters); 
	


	for(i = 0; i < data_size; i++)
		delete datamatrix[i];
	delete datamatrix;
			
	for(i = 0; i < all_dims; i++)
	{
		on[i] = on[map_file_to_buf[orderarray[i]]];
		map_buf_to_file[i] = orderarray[i];
	}
		
	for(i = 0; i < all_dims; i++)
		map_file_to_buf[map_buf_to_file[i]] = i;

    RemapAllData();

    if ( g_globals.dim_reducer != NULL )
    {
        delete g_globals.dim_reducer;
        g_globals.dim_reducer = NULL;
    }
}


// Added by Debbie
void okcdata::Clutter_Based_Order(char *display)  
{
	if(strcmp(display, "glyph") == 0)
	{

		SaveBrushes();

		int i,j;
		int set_size;
		double **norm_matrix;
		ARR<double> data(dims);

		norm_matrix = new double *[data_size];
		for(i = 0; i < data_size; i++)
			norm_matrix[i] = new double [dims];

		//param is defined in OrderDimension.h
		if (g_globals.reorderParameters->brushed == 0 )
		{
			for(i = 0; i < data_size; i++)
			{
				GetData(data,i);
				for(j = 0; j< dims; j++)
				{
					norm_matrix[i][j] = (data[map_file_to_buf[j]] - dim_min[map_file_to_buf[j]])/(dim_max[map_file_to_buf[j]] - dim_min[map_file_to_buf[j]]);
				}
			}
			set_size = data_size;
		}
		else 
		{
			set_size = 0;
			for(i = 0; i < data_size; i++)
			{
				GetData(data,i);
				if(BrushOperCheckPoint(data,0))
				{
					for(j = 0; j < dims; j++)
			    		norm_matrix[set_size][j] = (data[map_file_to_buf[j]] - dim_min[map_file_to_buf[j]])/(dim_max[map_file_to_buf[j]] - dim_min[map_file_to_buf[j]]);
					set_size++;
				}
			}
		}

		if( set_size <= 1)
		return;

		ARR<int> orderarray(all_dims);			// initialize the order array with the order in map_buf_to_file
		for(i = 0; i < all_dims; i++)
			orderarray[i] = i;

	

		clutterbased_reorder_glyph(set_size, dims, norm_matrix, orderarray, *g_globals.reorderParameters);

		for(i = 0; i < data_size; i++)
			delete norm_matrix[i];
		delete norm_matrix;
				
		for(i = 0; i < all_dims; i++)
		{
			on[i] = on[map_file_to_buf[orderarray[i]]];
			map_buf_to_file[i] = orderarray[i];
		}
			
		for(i = 0; i < all_dims; i++)
			map_file_to_buf[map_buf_to_file[i]] = i;

		RemapAllData();

		if ( g_globals.dim_reducer != NULL )
		{
			delete g_globals.dim_reducer;
			g_globals.dim_reducer = NULL;
		}
	}

	else if (strcmp(display, "dimstck") == 0)
	{
	SaveBrushes();

	int i,j;
	int set_size;
	double **datamatrix;
	ARR<double> data(dims);
	double temp;

	datamatrix = new double *[data_size];
	for(i = 0; i < data_size; i++)
		datamatrix[i] = new double [dims];

	if ( g_globals.reorderParameters->brushed == 0 )
	{
		for(i = 0; i < data_size; i++)
		{
			GetData(data,i);
			for(j = 0; j< dims; j++)
			{
				datamatrix[i][j] = data[map_file_to_buf[j]];
			}
		}
		set_size = data_size;
	}
	else 
	{
		set_size = 0;
		for(i = 0; i < data_size; i++)
		{
			GetData(data,i);
			if(BrushOperCheckPoint(data,0))
			{
				for(j = 0; j < dims; j++)
					temp = datamatrix[set_size][j] = data[map_file_to_buf[j]];
				set_size++;
			}
		}
		
	}
	if( set_size <= 1)
		return;

		ARR<int> orderarray(all_dims);			// initialize the order array with the order in map_buf_to_file
			for(i = 0; i < all_dims; i++)
				orderarray[i] = i;


		clutterbased_reorder_dimstck(set_size, dims, datamatrix, orderarray, *g_globals.reorderParameters);

		
		for(i = 0; i < all_dims; i++)
			{
				on[i] = on[map_file_to_buf[orderarray[i]]];
				map_buf_to_file[i] = orderarray[i];
			}
				
			for(i = 0; i < all_dims; i++)
				map_file_to_buf[map_buf_to_file[i]] = i;

			RemapAllData();

			if ( g_globals.dim_reducer != NULL )
			{
				delete g_globals.dim_reducer;
				g_globals.dim_reducer = NULL;
			}

	}
	

	else if (strcmp(display,"par")==0)
	{
		SaveBrushes();

		int i,j;
		int set_size;
		double **norm_matrix;
		ARR<double> data(dims);

		norm_matrix = new double *[data_size];
		for(i = 0; i < data_size; i++)
			norm_matrix[i] = new double [dims];

		if ( g_globals.reorderParameters->brushed == 0 )
		{
			for(i = 0; i < data_size; i++)
			{
				GetData(data,i);
				for(j = 0; j< dims; j++)
				{
					norm_matrix[i][j] = (data[map_file_to_buf[j]] - dim_min[map_file_to_buf[j]])/(dim_max[map_file_to_buf[j]] - dim_min[map_file_to_buf[j]]);
				}
			}
			set_size = data_size;
			

		}
		else 
		{
			set_size = 0;
			for(i = 0; i < data_size; i++)
			{
				GetData(data,i);
				if(BrushOperCheckPoint(data,0))
				{
					for(j = 0; j < dims; j++)
			    		norm_matrix[set_size][j] = (data[map_file_to_buf[j]] - dim_min[map_file_to_buf[j]])/(dim_max[map_file_to_buf[j]] - dim_min[map_file_to_buf[j]]);
					set_size++;
				}
			}
		}
	
		if( set_size <= 1)
		return;

		ARR<int> orderarray(all_dims);			// initialize the order array with the order in map_buf_to_file
			for(i = 0; i < all_dims; i++)
				orderarray[i] = i;


		clutterbased_reorder_parcord(set_size, dims, norm_matrix, orderarray, *g_globals.reorderParameters);
		
		for(i = 0; i < all_dims; i++)
			{
				on[i] = on[map_file_to_buf[orderarray[i]]];
				map_buf_to_file[i] = orderarray[i];
			}
				
			for(i = 0; i < all_dims; i++)
				map_file_to_buf[map_buf_to_file[i]] = i;

			RemapAllData();

			if ( g_globals.dim_reducer != NULL )
			{
				delete g_globals.dim_reducer;
				g_globals.dim_reducer = NULL;
			}

	}

	else if (strcmp(display, "scatter") == 0)
	{
		    SaveBrushes();

	int i,j;
	int set_size;
	double **datamatrix;
	ARR<double> data(dims);
	
	datamatrix = new double *[data_size];
	for(i = 0; i < data_size; i++)
		datamatrix[i] = new double [dims];

	if ( g_globals.reorderParameters->brushed == 0 )
	{
		for(i = 0; i < data_size; i++)
		{
			GetData(data,i);
			for(j = 0; j< dims; j++)
			{
				datamatrix[i][j] = data[map_file_to_buf[j]];
			}
		}
		set_size = data_size;
	}
	else 
	{
		set_size = 0;
		for(i = 0; i < data_size; i++)
		{
			GetData(data,i);
			if(BrushOperCheckPoint(data,0))
			{
				for(j = 0; j < dims; j++)
					datamatrix[set_size][j] = data[map_file_to_buf[j]];
				set_size++;
			}
		}
		
	}
	if( set_size <= 1)
		return;

		ARR<int> orderarray(all_dims);			// initialize the order array with the order in map_buf_to_file
			for(i = 0; i < all_dims; i++)
				orderarray[i] = i;


			clutterbased_reorder_scatterplot(set_size, dims, datamatrix, orderarray, *g_globals.reorderParameters);
		//Added by Zaixian begin
		//This code is to check whether orderarray is valid
		ARR<int> pos(all_dims);
		for (i=0; i<all_dims; i++)
			pos[i] = 0;
		for (i=0; i<all_dims; i++)
			pos[orderarray[i]] = 1;
		for (i=0; i<all_dims; i++)
			//Reset the orderarry if the orderarray is invalid
			if (pos[i]==0) {
				for (j=0; j<all_dims; j++)
					orderarray[j] = j;
				break;
			}
		//Zaixian end
		for(i = 0; i < all_dims; i++)
			{
				on[i] = on[map_file_to_buf[orderarray[i]]];
				map_buf_to_file[i] = orderarray[i];
			}
				
			for(i = 0; i < all_dims; i++)
				map_file_to_buf[map_buf_to_file[i]] = i;

			RemapAllData();

			if ( g_globals.dim_reducer != NULL )
			{
				delete g_globals.dim_reducer;
				g_globals.dim_reducer = NULL;
			}

	}
}


/*
void okcdata::SetReorderedBrushes()
{
	int br, i, loopi;
	ARR<double> temp_pos;
	ARR<double> temp_size;
	ARR<double> temp_bound_offset;  // boundary offset from min/max
	ARR<bool> temp_on;

	pos.SetLen(all_dims);
	size.SetLen(all_dims);
	temp_bound_offset.SetLen(all_dims);
	temp_on.SetLen(all_dims);

	for ( i = 0; i < all_dims; i++ )
	{
		temp_on[i] = on[i];
	}

	for ( loopi = 0; loopi < all_dims; loopi++ )
	{
		i = 
		temp_on[i] = on[i];
	}

	for (int br=0; br<MAXBRUSH; br++)
	{
		Brush *brush = &all_brushes[br];
		for ( i = 0; i < all_dims; i ++ )
		{
			temp_pos[i] =  brush.pos[i];
			temp_pos[i] =  brush.pos[i];
			temp_pos[i] =  brush.pos[i];
		}

		
	dims = all_dims;
	names.SetLen(all_dims);
	dim_min.SetLen(all_dims);
	dim_max.SetLen(all_dims);
	nominal_fields.SetLen(all_dims);
	org_dim_min.SetLen(all_dims);
	org_dim_max.SetLen(all_dims);

	cardinality.SetLen(all_dims);
	tempCardinality.SetLen(all_dims);

	dims = 0;
	for (i=0;i<all_dims;i++)
		if (on[i]) {
			org_dim_min[dims] = dim_min[dims] = all_dim_min[i];
			org_dim_max[dims] = dim_max[dims] = all_dim_max[i];
			cardinality[dims] = all_cardinality[i];
			names[dims] = all_names[i];
			nominal_fields[dims] = all_nominal_fields[i];
			dims++;
		}


	one_over_dims = 1.0/dims;

	for (br=0; br<MAXBRUSH; br++)
	{
		int j = 0;
		for (int i=0;i<MIN2(brushes[br].pos.Len(),on.Len());i++)
			if (on[i]) {
				brushes[br].pos[j] = all_brushes[br].pos[i];
				brushes[br].size[j] = all_brushes[br].size[i];
				brushes[br].bound_offset[j] = all_brushes[br].bound_offset[i];
				j++;
			}
	}
}
*/

const char * NominalField::MapValueToString(int what)
{
	if (what<range_min || what>range_max || values[what - range_min] == 0) {
		static char number_string[10];
		sprintf(number_string, "%d", what);
		return number_string;
	}
	return values[what - range_min];
}


void okcdata::MapDataWithDimClusterExtends(ARR<double> all_data,
										   ARR<double> &mean,
										   ARR<double> &max,
										   ARR<double> &min)
{
	// PAUL: lang_opt
	// temp is static, so it does not have to be reallocated
	static ARR<double> temp;
	temp.SetLen(all_dims);
	int i, loopi;
	for (loopi = 0; loopi < all_dims; loopi++) {
		i = map_file_to_buf[loopi];
		temp[i] = all_data[loopi];
	}

	// PAUL: lang_opt
	static ARR<double> temp1;
	temp1.SetLen(all_dims);
	int count = 0;
	for (i = 0; i < all_dims; i++) {
		if (on[i]) {
			temp1[count] = temp[i];
			count++;
		}
	}

	g_globals.dim_reducer->MapDataWithDimClusterExtends(temp1, mean, max, min);
}


/*
template <class T> void okcdata::MapDataWithoutOnOff( ARR<T> all_data, ARR<T> &data)
{
	int i, loopi;
	for (loopi=0; loopi<all_dims; loopi++)
	{
		i = map_file_to_buf[loopi];
		data[i] = all_data[loopi];
	}

}
*/

//shiping
void okcdata::setNumberOfBins(int which, int num)
{
	tempCardinality[which] = num;
}
//shiping

void okcdata::setWeightOfDim(int which, int num)
{
	tempWeight[which] = num;
}

// Anil [5/14/2004]
static inline double norm(int dim, double val)
{
	return (val - g_globals.dataset->dim_min[dim]) / 
			(g_globals.dataset->dim_max[dim] - g_globals.dataset->dim_min[dim]);
}
//

//Anil
const char * okcdata::MapValueToString(int dim, double value) const
{
	static char caption[100];
	memset((void *) caption, 0, 100 * sizeof(char));

	int valueTemp = (int) (value + 0.5);

#ifdef USE_META
	if (hasMeta) {
		//std::cerr << "MapValueToString: " << dim << std::endl;
		IDimension *dimension = meta->getDimension(dim);
		if (dimension) {
			Type dimType = dimension->getType();
			assert(!dimType.isBad());

			if (dimType.isDerivedFrom(Numeric::getTypeId())) {
				if (dimType == Integer::getTypeId()) {
					sprintf(caption, "%d", (int) value);
					return caption;
				}

				if (dimType == Real::getTypeId()) {
					sprintf(caption, "%.02f", value);
					return caption;
				}
			} else {
				IValue *mappedValue = dimension->getMappedValue((float)
																value);
				if (mappedValue) {
					std::string strValue = mappedValue->getStringRepresentation();
					if (strValue.length()) {
						sprintf(caption, "%s", strValue.c_str());
					}
				}
			}
		}
	} else 
		#endif // USE_META
	{
		if (nominal_fields[dim] != 0) {
			const char *string = nominal_fields[dim]->MapValueToString(valueTemp);
			sprintf(caption, "%s", string);
		} else {
			sprintf(caption, "%.2lf", value);
		}
	}
	return caption;
}

const char * okcdata::GetMin(int dim)
{
	static char buf[100] = "";

#ifdef USE_META
	if (hasMeta) {
		Type dimType;
		IDimension *dimension = meta->getDimension(GetRealPos(dim));
		if (dimension) {
			dimType = dimension->getType();
		}

		assert(!dimType.isBad());

		if (!(dimType.isDerivedFrom(Numeric::getTypeId()))) {
			return "";
		}

		if (dimType == Integer::getTypeId()) {
			sprintf(buf, "%d", (int) dim_min[dim]);
			return buf;
		}

		if (dimType == Real::getTypeId()) {
			sprintf(buf, "%.02f", dim_min[dim]);
			return buf;
		}
	} else 
		#endif // USE_META
	{
		sprintf(buf, "%.02f", dim_min[dim]);
		return buf;
	}

	// this should not occur
	return "min not defined";
}

const char * okcdata::GetMax(int dim)
{
	static char buf[100] = "";

#ifdef USE_META
	if (hasMeta) {
		Type dimType;
		IDimension *dimension = meta->getDimension(GetRealPos(dim));
		if (dimension) {
			dimType = dimension->getType();
		}

		assert(!dimType.isBad());

		if (!(dimType.isDerivedFrom(Numeric::getTypeId()))) {
			// nothing to be displayed for non-numeric values
			return "";
		}

		if (dimType == Integer::getTypeId()) {
			sprintf(buf, "%d", (int) dim_max[dim]);
			return buf;
		}

		if (dimType == Real::getTypeId()) {
			sprintf(buf, "%.02f", dim_max[dim]);
			return buf;
		}
	} else 
#endif // USE_META
	{
		sprintf(buf, "%.02f", dim_max[dim]);
		return buf;
	}

	// this should not occur
	return "max not defined";
}

#ifdef USE_META
IMeta * okcdata::GetMeta() const
{
	if (hasMeta) {
		return meta;
	}
	return 0;
}

void okcdata::ShowContents(Tcl_Interp *interp, int dim, double value) const
{
	if (!hasMeta) {
		return;
	}

	Type dimType;
	IDimension *dimension = meta->getDimension(dim);
	if (dimension) {
		dimType = dimension->getType();
	}

	// this should be a valid type
	assert(!dimType.isBad());

	if (dimType == URI::getTypeId()) {
		IValue *mappedValue = dimension->getMappedValue((float) value);

		if (mappedValue) {
			appShowURIDialog(interp,
							 mappedValue->getStringRepresentation(),
							 mappedValue->getStringContents());
		}
	}
}

void okcdata::CreateNewMeta()
{
	assert(!hasMeta);	// make sure we do not have a meta yet!

	meta = MetaFactory::instance()->createMeta();
	meta->setDataPointCount(data_size);
	meta->setDimensionCount(all_dims);
	for (int i = 0; i < all_dims; ++i) {
		meta->addDimension(meta->createDimension(all_names[i],
												 Real::getTypeId()));
	}

	hasMeta = true;
}
#endif // USE_META
//Anil

//added by Zaixian
int okcdata::GetRealPos(int dim)
{
	int posInBuf=0, count = 1;

	while (posInBuf<all_dims) {
		while ( (posInBuf<all_dims) && (!on[posInBuf]) ) posInBuf++;
		if (posInBuf>=all_dims) return -1;
		if (count==dim+1) return map_buf_to_file[posInBuf];
		posInBuf++;
		count++;
	}

	return -1;

}
