#include <RegionFactory.h>
#include <CacheEntry.h>
#include <RemainderRegion.h>
#include <Brush.h>
#include <vector>
#include <list>
#include <CacheRegion.h>
#include <ICacheEntry.h>

using namespace cache;
using namespace std;

std::vector<RemainderRegion*>& AbstractRegionFactory :: 
makeRegions(std::list<ICacheEntry*>& listCacheObjects,
			Brush& brush)
{	
	Place place;
	this->removeRegions(this->vecRemainderRegionPtr);

	//Making the border regions
	RemainderRegion *remainderRegion= NULL;

	if(listCacheObjects.size() == 0)//There is nothing in the cache 
	{
		place = both;
		remainderRegion = make(brush.getLeftExtent(),
			brush.getRightExtent(),
			brush.getLevelOfDetail(),
			brush.getBrushType(),
			place);
		vecRemainderRegionPtr->push_back(remainderRegion);
		return(*this->vecRemainderRegionPtr);
	} // if(numberOfRegions == 0)//There is nothing in the cache

	std::vector<CacheRegion> vecCacheRegion;
	//Makes the cache Regions and sorts them
	makeCacheRegions(vecCacheRegion,listCacheObjects);

	//Clear the vector of Remainder Regions
	this->makeInBetweenRegions(*vecRemainderRegionPtr,
		vecCacheRegion,
		brush.getLevelOfDetail());

	//Check to see if there are remainder regions at the border of the brush
	if(brush.getLeftExtent() < 
		vecCacheRegion[0].getLeftExtent())
	{
		place = left;
		//Note we make the left and right remainder regions to avoid duplicates.
		//That is we in case of the left remainder region we make sure that the X < e2.
		//Because if we don't do that.
		//That is we keep the fetch condition to X <= e2 we will also the fetch 1st cache region.		
		remainderRegion = make(
			brush.getLeftExtent(),
			vecCacheRegion[0].getLeftExtent(),
			brush.getLevelOfDetail(),
			brush.getBrushType(),
			place);

		//Note you are pushing the left region also at the end			
		vecRemainderRegionPtr->insert(vecRemainderRegionPtr->begin(),
			remainderRegion);
	}

	int size = vecCacheRegion.size();
	if(vecCacheRegion[size-1].getRightExtent() 
		< brush.getRightExtent())
	{		
		place = right;
		//The same reason above we make sure that Y > e1 because if we don't do that we will also fetch the 
		//last cache region from the data base.
		//Make Right region
		remainderRegion = 
			make(
			vecCacheRegion[size-1].getRightExtent(),
			brush.getRightExtent(),
			brush.getLevelOfDetail(),
			brush.getBrushType(),
			place);
		vecRemainderRegionPtr->push_back(remainderRegion);
	}		
	return(*vecRemainderRegionPtr);	
}		

AbstractRegionFactory ::  ~AbstractRegionFactory()
{
	this->removeRegions(this->vecRemainderRegionPtr);
	delete(this->vecRemainderRegionPtr);		
	vecRemainderRegionPtr = NULL;
}


void AbstractRegionFactory ::  makeInBetweenRegions(
	std::vector<RemainderRegion*> &vecRemainderRegion,
	std::vector<CacheRegion>& vecCacheRegion,
	double levelOfDetail)
{
	RemainderRegion *remainderRegion=NULL;

	for(int i=0;i< (int) vecCacheRegion.size()-1;i++)
	{			
		remainderRegion = 
			fillGap(vecCacheRegion[i],
			vecCacheRegion[i+1],
			levelOfDetail);

		if(remainderRegion != NULL)		
			vecRemainderRegion.push_back(remainderRegion);
	} // for(unsigned int i=0;i<vecCacheRegion.size()-1;i++)		
}

void AbstractRegionFactory ::  makeCacheRegions(std::vector<CacheRegion>& vecCacheRegion,		
												std::list<ICacheEntry*>& listCacheObjects)
{
	std::list<ICacheEntry*>::iterator c_iterator;
	c_iterator = listCacheObjects.begin();

	for(unsigned int count = 0; count < listCacheObjects.size();count++,c_iterator++)				
	{	
		CacheRegion cacheRegion(**c_iterator);
		vecCacheRegion.push_back(cacheRegion);			
	}
	//Sorting the cacheRegions based on the E1 extents
	//vector<CacheRegion>::iterator first =  vecCacheRegion.begin();
	//vector<CacheRegion>::iterator last  =  vecCacheRegion.end();
	//std::sort(first, last, cache::CacheRegion::operator<);

	std::sort(vecCacheRegion.begin(),vecCacheRegion.end(), compare);		

}

RemainderRegion* AbstractRegionFactory ::  make(double leftExtent,
												double rightExtent,
												double levelOfDetail,
												Brush::Type type,
												Place place)
{
	RemainderRegion* region = NULL;
	if(type == 0)
	{
		if(place == left)
			region = 
			new LeftAnyRemainderRegion(leftExtent,
			rightExtent,
			levelOfDetail);
		if(place == right)
			region = 
			new RightAnyRemainderRegion(leftExtent,
			rightExtent,
			levelOfDetail);
		if(place == both)
			region = 
			new AnyRemainderRegion(leftExtent,
			rightExtent,
			levelOfDetail);
	}
	if(type == 1)
	{
		region = new AllRemainderRegion(leftExtent,
			rightExtent,
			levelOfDetail);
	}
	return(region);
}

void AbstractRegionFactory ::  removeRegions(std::vector<RemainderRegion*> *vecRemainderRegion)
{
	//Delete all the remainder regions
	int size = vecRemainderRegion->size();
	for(int i=size;i>0;i--)
	{
		RemainderRegion* region = (*vecRemainderRegion)[i-1];
		vecRemainderRegion->pop_back();		
		delete(region);			
	}
}


/*
*	A singleton class makes the remainder regions
* Uses the class cache Region and the frinend function defined in
* CacheRegion.h
*/


AscRegionFactory ::AscRegionFactory()
{
	this->vecRemainderRegionPtr = new std::vector<RemainderRegion*>;
}
/**
*
* \return RegionFactory<T>*
*/
AscRegionFactory* AscRegionFactory :: getInstance()
{
	if(regionFactoryPtr == NULL)
		regionFactoryPtr = new AscRegionFactory();
	return(regionFactoryPtr);		
}

AscRegionFactory* &AscRegionFactory::getReference()
{
	return(regionFactoryPtr);
}

//Defining the static variable
AscRegionFactory* AscRegionFactory::regionFactoryPtr;

/**Note this might not be the most efficient implementation but
* But this is not my focus right now.
*/

DescRegionFactory ::DescRegionFactory()
{
	this->vecRemainderRegionPtr = new std::vector<RemainderRegion*>;
}

/**	
* \param list 
* \param brush 
* \return 
*/
std::vector<RemainderRegion*>& DescRegionFactory :: makeRegions(
	std::list<ICacheEntry*>& listCacheObjects,
	Brush& brush)
{

	AbstractRegionFactory::makeRegions(listCacheObjects,brush);

	std::vector<RemainderRegion*> tempVector;	

	for(int i=vecRemainderRegionPtr->size()-1;i>=0 ;i--)
		tempVector.push_back((*vecRemainderRegionPtr)[i]);		

	vecRemainderRegionPtr->clear();

	for(unsigned int i=0;i<tempVector.size();i++)
		vecRemainderRegionPtr->push_back(tempVector[i]);

	return(*vecRemainderRegionPtr);
}


/**
*
* \return RegionFactory<T>*
*/
DescRegionFactory* DescRegionFactory :: getInstance()
{
	if(regionFactoryPtr == NULL)
		regionFactoryPtr = new DescRegionFactory();
	return(regionFactoryPtr);		
}

DescRegionFactory*& DescRegionFactory::getReference()
{
	return(regionFactoryPtr);
}

//Defining the static variable
DescRegionFactory* DescRegionFactory::regionFactoryPtr;
