
#include <cache/Cache.h>
#include <cache/PerformanceStatistics.h>
#include <cache/SuperIndex.h>

using namespace cache;

SuperIndex::SuperIndex(unsigned long sizeOfCache,
							unsigned int switchThreshold,	
							PerformanceStatistics* perStatistics,
							SequentialIndex* seqIndex,
							RTreeIndex* rtreeIndex)
{
	this->switchThreshold = switchThreshold;
	this->rTreeIndex = rtreeIndex;
	this->seqIndex = seqIndex;
	this->sizeOfCache = sizeOfCache;
	this->perStatistics = perStatistics;

}
		
	/**
	* Deletes the data items
	* \param *iShape 
	* \param id 
	*/
	void SuperIndex::deleteData(Tools::Geometry::IShape *iShape,unsigned long id)
	{
		this->rTreeIndex->deleteData(iShape,id);
		this->seqIndex->deleteData(iShape,id);

	}

		/**
		* Inserts the cache entry into the index.
		* Called when a new 
		* \param &cacheEntry 
		* \return 
		*/
	void SuperIndex::insertData(ICacheEntry& cacheEntry)
	{
		this->rTreeIndex->insertData(cacheEntry);
		this->seqIndex->insertData(cacheEntry);
	}

		/*
		*	Destroys the sequential index
		*/
	SuperIndex::~SuperIndex()
	{
		delete(this->rTreeIndex);
		delete(this->seqIndex);
	}

	/**
	* Does a sequential scan over the complete data set in the main memory
	* \param &brush 
	* \return 
	*/
	std::list<ICacheEntry*>& SuperIndex::search(Brush &brush)
	{
				
		double descParam = (double)(this->perStatistics->brushSize)/(double)this->sizeOfCache * 100;

		if(descParam > this->switchThreshold)		
			return(this->seqIndex->search(brush));
		
		else		
			return(this->rTreeIndex->search(brush));		
		
	}
