// MetaHelper.cpp
//
//////////////////////////////////////////////////////////////////////

#pragma warning(disable:4786)

#include <iostream>
#include <SAX/SAXException.h>
#include <meta/MetaHelper.h>

/*static*/ MetaHelper* MetaHelper::global = 0;

//////////////////////////////////////////////////////////////////////

class MetaErrorHandler : public SAX::ErrorHandler
{
public:
	static MetaErrorHandler *instance() {
		static MetaErrorHandler *error = 0;
		if (!error) {
			error = new MetaErrorHandler();
		}
		return error;
	}

	void warning(const SAX::SAXParseException& exception);
	void error(const SAX::SAXParseException& exception);
	void fatalError(const SAX::SAXParseException& exception);
};

//////////////////////////////////////////////////////////////////////
// MetaHelper

MetaHelper::MetaHelper()
{
	parser.setErrorHandler(*MetaErrorHandler::instance());
}

void MetaHelper::pushHandler(SAX::ContentHandler *handler)
{
	contentHandlers.push(handler);
	parser.setContentHandler(*handler);
}

SAX::ContentHandler* MetaHelper::popHandler()
{
	SAX::ContentHandler* handler = contentHandlers.top();
	contentHandlers.pop();
	if (!contentHandlers.empty()) {
		parser.setContentHandler(*contentHandlers.top());
	}
	return handler;
}

SAX::XMLReader<std::string>& MetaHelper::getParser()
{
	return parser;
}

/*static*/ void MetaHelper::reinitialize()
{
	while(!global->contentHandlers.empty())
		global->contentHandlers.pop();
}

//////////////////////////////////////////////////////////////////////
// MetaErrorHandler
void MetaErrorHandler::warning(const SAX::SAXParseException& exception)
{
	std::cerr << exception.what() << std::endl;
}

void MetaErrorHandler::error(const SAX::SAXParseException& exception)
{
	std::cerr << exception.what() << std::endl;
}

void MetaErrorHandler::fatalError(const SAX::SAXParseException& exception)
{
	std::cerr << exception.what() << std::endl;
}
